# -*- coding: utf-8 -*-
"""
Created on Thu Apr 18 17:27:06 2024

@author: rensv
"""
# import dependencies
import numpy as np
from .structuralelement import StructuralElement, ElementFactory

# %% class definition


@ElementFactory.ElementType('EulerBernoulli Beam with foundation')
class EulerBernoulliBeamFoundation(StructuralElement):
    """Class for Euler-Bernoulli beam element over viscoelastic foundation."""


    def __init__(self, rho, A, E, Ib, L, k, ksi=None):
        """
        Initialise EB beam over viscoelastic foundation.

        Input:
            rho : value. Density of element [kg/m^3]
            A   : value. Area of cross section [m^2]
            E   : value. Young's modulus [Pa]
            Ib  : value. Cross-sectional inertia [m^4]
            L   : value. Length of the element [m]
            k   : value. Stiffness of viscoelastic foundation [N/m^2]
            ksi : value. Material damping [-], default: none
        """
        # define the of dofs the EulerBernoulli beam and initialise
        dofs = ['z','phi_y']
        super().__init__(dofs)

        # Initialise local beam element with necessary parameters
        self.rho = rho
        self.A = A
        self.E = E
        self.Ib = Ib
        self.L = L
        self.k = k
        # Assign ksi if given,  otherwise assign a default value
        self.ksi = ksi if ksi is not None else 0.01

    def LocalStiffness(self, omega):
        """
        Determine the stiffness of the EB-beam over viscoelastic foundation.

        As its 2D, stiffness will be 4x4 matrix as:
            [V_left, M_left, V_right, M_right] =
            K.[w_left, phi_left, w_right, phi_right]
        where:
            K = [K_V_ll, K_M_ll, K_V_lr, K_M_lr;
                 K_V_rl, K_M_rl, K_V_rr, K_M_rr]
        Input:
            omega: array. Range of frequencies of analysis
        Output:
            K_local: matrix. Dynamic stiffness matrix (also K_dyn)
        """
        # Assign local variables for ease of coding
        E = self.E
        Ib = self.Ib
        L = self.L
        # Obtain the wavenumbers
        alpha_1, alpha_2, alpha_3, alpha_4 = self.ElementWaveNumbers(omega)

        # Initialise K_local
        K_local = np.empty((4, 4), complex)
        # Copy-paste the matrix directly from Maple (also called K_dyn)
        K_local = np.array([[-1j*Ib*E*(alpha_3*alpha_4*(alpha_3 - alpha_4)*(alpha_3 + alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - alpha_2*alpha_4*(alpha_2 - alpha_4)*(alpha_2 + alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + alpha_2*alpha_3*(alpha_2 - alpha_3)*(alpha_2 + alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + alpha_1*(alpha_4*(alpha_1 - alpha_4)*(alpha_1 + alpha_4)*(alpha_2 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_3)) - alpha_3*(alpha_1 - alpha_3)*(alpha_1 + alpha_3)*(alpha_2 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*alpha_2*(alpha_1 - alpha_2)*(alpha_1 + alpha_2)*(alpha_3 - alpha_4)))/((alpha_3 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*(alpha_3 - alpha_4)*(alpha_1 - alpha_2)),
                             Ib*E*(((alpha_3**3 - alpha_4**3)*alpha_1 + (-alpha_3**3 + alpha_4**3)*alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) + ((-alpha_2**3 + alpha_4**3)*alpha_1 + alpha_3*alpha_2**3 - alpha_4**3*alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + ((alpha_2**3 - alpha_3**3)*alpha_1 - alpha_4*alpha_2**3 + alpha_4*alpha_3**3)*np.exp(-1j*L*(alpha_1 + alpha_4)) + ((alpha_2 - alpha_3)*alpha_1**3 - alpha_4**3*(alpha_2 - alpha_3))*np.exp(-1j*L*(alpha_2 + alpha_3)) + ((-alpha_2 + alpha_4)*alpha_1**3 + alpha_3**3*(alpha_2 - alpha_4))*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*(alpha_1 - alpha_2)*(alpha_1**2 + alpha_2*alpha_1 + alpha_2**2)*(alpha_3 - alpha_4))/((alpha_3 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*(alpha_3 - alpha_4)*(alpha_1 - alpha_2)),
                             1j*(alpha_1*(alpha_3 - alpha_4)*(alpha_2 - alpha_4)*(alpha_2 - alpha_3)*(alpha_2 + alpha_3 + alpha_4)*np.exp(-1j*alpha_1*L) - alpha_2*(alpha_3 - alpha_4)*(alpha_1 - alpha_4)*(alpha_1 - alpha_3)*(alpha_1 + alpha_3 + alpha_4)*np.exp(-1j*alpha_2*L) + (alpha_3*(alpha_2 - alpha_4)*(alpha_1 - alpha_4)*(alpha_1 + alpha_2 + alpha_4)*np.exp(-1j*alpha_3*L) - np.exp(-1j*alpha_4*L)*alpha_4*(alpha_2 - alpha_3)*(alpha_1 - alpha_3)*(alpha_1 + alpha_2 + alpha_3))*(alpha_1 - alpha_2))*Ib*E/((alpha_3 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*(alpha_3 - alpha_4)*(alpha_1 - alpha_2)),
                             -Ib*E*((alpha_3 - alpha_4)*(alpha_2 - alpha_4)*(alpha_2 - alpha_3)*(alpha_2 + alpha_3 + alpha_4)*np.exp(-1j*alpha_1*L) - (alpha_3 - alpha_4)*(alpha_1 - alpha_4)*(alpha_1 - alpha_3)*(alpha_1 + alpha_3 + alpha_4)*np.exp(-1j*alpha_2*L) + ((alpha_2 - alpha_4)*(alpha_1 - alpha_4)*(alpha_1 + alpha_2 + alpha_4)*np.exp(-1j*alpha_3*L) - np.exp(-1j*alpha_4*L)*(alpha_2 - alpha_3)*(alpha_1 - alpha_3)*(alpha_1 + alpha_2 + alpha_3))*(alpha_1 - alpha_2))/((alpha_3 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*(alpha_3 - alpha_4)*(alpha_1 - alpha_2))],
                            [Ib*E*(alpha_3*alpha_4*(alpha_3 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - alpha_2*alpha_4*(alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + alpha_2*alpha_3*(alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + alpha_1*(alpha_4*(alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - alpha_3*(alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*alpha_2*(alpha_3 - alpha_4)*(alpha_1 - alpha_2)))/((alpha_3 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*(alpha_3 - alpha_4)*(alpha_1 - alpha_2)),
                             1j*Ib*E*((alpha_3 - alpha_4)*(alpha_3 + alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - (alpha_2 - alpha_4)*(alpha_2 + alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + (alpha_2 - alpha_3)*(alpha_2 + alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + (alpha_1 - alpha_4)*(alpha_1 + alpha_4)*(alpha_2 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_3)) - (alpha_1 - alpha_3)*(alpha_1 + alpha_3)*(alpha_2 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*(alpha_1 - alpha_2)*(alpha_1 + alpha_2)*(alpha_3 - alpha_4))/((alpha_3 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*(alpha_3 - alpha_4)*(alpha_1 - alpha_2)),
                             -(alpha_1*(alpha_3 - alpha_4)*(alpha_2 - alpha_4)*(alpha_2 - alpha_3)*np.exp(-1j*alpha_1*L) - alpha_2*(alpha_3 - alpha_4)*(alpha_1 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*alpha_2*L) + (alpha_1 - alpha_2)*(alpha_3*(alpha_2 - alpha_4)*(alpha_1 - alpha_4)*np.exp(-1j*alpha_3*L) - np.exp(-1j*alpha_4*L)*alpha_4*(alpha_2 - alpha_3)*(alpha_1 - alpha_3)))*Ib*E/((alpha_3 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*(alpha_3 - alpha_4)*(alpha_1 - alpha_2)),
                             -1j*Ib*E*((alpha_3 - alpha_4)*(alpha_2 - alpha_4)*(alpha_2 - alpha_3)*np.exp(-1j*alpha_1*L) - (alpha_3 - alpha_4)*(alpha_1 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*alpha_2*L) + ((alpha_2 - alpha_4)*(alpha_1 - alpha_4)*np.exp(-1j*alpha_3*L) - np.exp(-1j*alpha_4*L)*(alpha_2 - alpha_3)*(alpha_1 - alpha_3))*(alpha_1 - alpha_2))/((alpha_3 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*(alpha_3 - alpha_4)*(alpha_1 - alpha_2))],
                            [1j*(alpha_4*(alpha_2 - alpha_3)*(alpha_1 - alpha_3)*(alpha_1 - alpha_2)*(alpha_1 + alpha_2 + alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_2 + alpha_3)) - alpha_3*(alpha_2 - alpha_4)*(alpha_1 - alpha_4)*(alpha_1 - alpha_2)*(alpha_1 + alpha_2 + alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_2 + alpha_4)) + (alpha_3 - alpha_4)*(alpha_2*(alpha_1 - alpha_4)*(alpha_1 - alpha_3)*(alpha_1 + alpha_3 + alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_3 + alpha_4)) - np.exp(-1j*L*(alpha_2 + alpha_3 + alpha_4))*alpha_1*(alpha_2 - alpha_4)*(alpha_2 - alpha_3)*(alpha_2 + alpha_3 + alpha_4)))*Ib*E/((alpha_3 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*(alpha_3 - alpha_4)*(alpha_1 - alpha_2)),
                             -Ib*E*((alpha_2 - alpha_3)*(alpha_1 - alpha_3)*(alpha_1 - alpha_2)*(alpha_1 + alpha_2 + alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_2 + alpha_3)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_4)*(alpha_1 - alpha_2)*(alpha_1 + alpha_2 + alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_2 + alpha_4)) + (alpha_3 - alpha_4)*((alpha_1 - alpha_4)*(alpha_1 - alpha_3)*(alpha_1 + alpha_3 + alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_3 + alpha_4)) - np.exp(-1j*L*(alpha_2 + alpha_3 + alpha_4))*(alpha_2 - alpha_4)*(alpha_2 - alpha_3)*(alpha_2 + alpha_3 + alpha_4)))/((alpha_3 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*(alpha_3 - alpha_4)*(alpha_1 - alpha_2)),
                             1j*Ib*E*(alpha_1*alpha_2*(alpha_1 - alpha_2)*(alpha_1 + alpha_2)*(alpha_3 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_2)) - alpha_1*alpha_3*(alpha_1 - alpha_3)*(alpha_1 + alpha_3)*(alpha_2 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_3)) + alpha_1*alpha_4*(alpha_1 - alpha_4)*(alpha_1 + alpha_4)*(alpha_2 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_4)) + alpha_2*alpha_3*(alpha_2 - alpha_3)*(alpha_2 + alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - alpha_4*(alpha_2*(alpha_2 - alpha_4)*(alpha_2 + alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) - np.exp(-1j*L*(alpha_3 + alpha_4))*alpha_3*(alpha_3 - alpha_4)*(alpha_3 + alpha_4)*(alpha_1 - alpha_2)))/((alpha_3 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*(alpha_3 - alpha_4)*(alpha_1 - alpha_2)),
                             -(((alpha_3 - alpha_4)*alpha_1**3 - alpha_2**3*(alpha_3 - alpha_4))*np.exp(-1j*L*(alpha_1 + alpha_2)) + ((-alpha_2 + alpha_4)*alpha_1**3 + alpha_3**3*(alpha_2 - alpha_4))*np.exp(-1j*L*(alpha_1 + alpha_3)) + ((alpha_2 - alpha_3)*alpha_1**3 - alpha_4**3*(alpha_2 - alpha_3))*np.exp(-1j*L*(alpha_1 + alpha_4)) + ((alpha_2**3 - alpha_3**3)*alpha_1 - alpha_4*alpha_2**3 + alpha_4*alpha_3**3)*np.exp(-1j*L*(alpha_2 + alpha_3)) + ((-alpha_2**3 + alpha_4**3)*alpha_1 + alpha_3*alpha_2**3 - alpha_4**3*alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*(alpha_3 - alpha_4)*(alpha_3**2 + alpha_3*alpha_4 + alpha_4**2)*(alpha_1 - alpha_2))*Ib*E/((alpha_3 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*(alpha_3 - alpha_4)*(alpha_1 - alpha_2))],
                            [-(alpha_4*(alpha_2 - alpha_3)*(alpha_1 - alpha_3)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2 + alpha_3)) - alpha_3*(alpha_2 - alpha_4)*(alpha_1 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2 + alpha_4)) + (alpha_3 - alpha_4)*(alpha_2*(alpha_1 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3 + alpha_4)) - np.exp(-1j*L*(alpha_2 + alpha_3 + alpha_4))*alpha_1*(alpha_2 - alpha_4)*(alpha_2 - alpha_3)))*Ib*E/((alpha_3 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*(alpha_3 - alpha_4)*(alpha_1 - alpha_2)),
                             -1j*Ib*((alpha_2 - alpha_3)*(alpha_1 - alpha_3)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2 + alpha_3)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2 + alpha_4)) + (alpha_3 - alpha_4)*((alpha_1 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3 + alpha_4)) - np.exp(-1j*L*(alpha_2 + alpha_3 + alpha_4))*(alpha_2 - alpha_4)*(alpha_2 - alpha_3)))*E/((alpha_3 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*(alpha_3 - alpha_4)*(alpha_1 - alpha_2)),
                             -(alpha_1*alpha_2*(alpha_3 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - alpha_1*alpha_3*(alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + alpha_1*alpha_4*(alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + alpha_2*alpha_3*(alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - alpha_4*(alpha_2*(alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) - np.exp(-1j*L*(alpha_3 + alpha_4))*alpha_3*(alpha_3 - alpha_4)*(alpha_1 - alpha_2)))*Ib*E/((alpha_3 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*(alpha_3 - alpha_4)*(alpha_1 - alpha_2)),                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                               -1j*((alpha_1 - alpha_2)*(alpha_1 + alpha_2)*(alpha_3 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_2)) - (alpha_1 - alpha_3)*(alpha_1 + alpha_3)*(alpha_2 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_3)) + (alpha_1 - alpha_4)*(alpha_1 + alpha_4)*(alpha_2 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_4)) + (alpha_2 - alpha_3)*(alpha_2 + alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - (alpha_2 - alpha_4)*(alpha_2 + alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*(alpha_3 - alpha_4)*(alpha_3 + alpha_4)*(alpha_1 - alpha_2))*Ib*E/((alpha_3 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*(alpha_3 - alpha_4)*(alpha_1 - alpha_2))]])
        
        return K_local

    def ElementWaveNumbers(self, omega):
        """
        Determine the wavenumbers of EB beam over viscoelastic foundation.

        Input:
            omega: array. Range of frequencies of analysis.
        Output:
            alpha_1, alpha_2, alpha_3, alpha_4: values. Wavenumbers
        """
        rho = self.rho
        A = self.A
        E = self.E
        Ib = self.Ib
        k = self.k
        # Copy-paste from Maple document of Wavenumbers
        # 4th-order derivative in x, so 4 wavenumbers
        alpha_1 = 1/E/Ib*((A*omega**2*rho - k)*E**3*Ib**3)**(1/4)
        alpha_2 = 1j/E/Ib*((A*omega**2*rho - k)*E**3*Ib**3)**(1/4)
        alpha_3 = -1/E/Ib*((A*omega**2*rho - k)*E**3*Ib**3)**(1/4)
        alpha_4 = -1j/E/Ib*((A*omega**2*rho - k)*E**3*Ib**3)**(1/4)

        return alpha_1, alpha_2, alpha_3, alpha_4

    def LocalDistributedLoad(self, q, omega):
        """
        Add a distributed load to the local element.

        Input:
            q: array. Distributed load. With a shape such as:
            q = [q_z, q_phi]
            omega: array. Range of frequencies of analyis
        Output:
            el: array. Force vector.
        """
        # assign load to itself to keep track
        self.q = q

        # assign local variables for ease of coding
        rho = self.rho
        A = self.A
        E = self.E
        Ib = self.Ib
        L = self.L
        k = self.k

        # determine wavenumber
        alpha_1, alpha_2, alpha_3, alpha_4 = self.ElementWaveNumbers(omega)

        # extract loads
        q_z = q[0]
        q_phi = q[1]

        # TODO - check for correctness
        el = np.array([1j*Ib*(alpha_3*alpha_4*(alpha_3 - alpha_4)*(alpha_3 + alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - alpha_2*alpha_4*(alpha_2 - alpha_4)*(alpha_2 + alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + alpha_2*alpha_3*(alpha_2 - alpha_3)*(alpha_2 + alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + (alpha_4*(alpha_1 - alpha_4)*(alpha_1 + alpha_4)*(alpha_2 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_3)) - alpha_3*(alpha_1 - alpha_3)*(alpha_1 + alpha_3)*(alpha_2 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*alpha_2*(alpha_1 - alpha_2)*(alpha_1 + alpha_2)*(alpha_3 - alpha_4))*alpha_1)*E/((alpha_3 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*(alpha_3 - alpha_4)*(alpha_1 - alpha_2))*q_z/(A*omega**2*rho - k) + -1j*(alpha_1*(alpha_3 - alpha_4)*(alpha_2 - alpha_4)*(alpha_2 - alpha_3)*(alpha_2 + alpha_3 + alpha_4)*np.exp(-1j*alpha_1*L) - alpha_2*(alpha_3 - alpha_4)*(alpha_1 - alpha_4)*(alpha_1 - alpha_3)*(alpha_1 + alpha_3 + alpha_4)*np.exp(-1j*alpha_2*L) + (alpha_3*(alpha_2 - alpha_4)*(alpha_1 - alpha_4)*(alpha_1 + alpha_2 + alpha_4)*np.exp(-1j*alpha_3*L) - np.exp(-1j*alpha_4*L)*alpha_4*(alpha_2 - alpha_3)*(alpha_1 - alpha_3)*(alpha_1 + alpha_2 + alpha_3))*(alpha_1 - alpha_2))*Ib*E/((alpha_3 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*(alpha_3 - alpha_4)*(alpha_1 - alpha_2))*q_z/(A*omega**2*rho - k),
                       -Ib*E*(alpha_3*alpha_4*(alpha_3 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - alpha_2*alpha_4*(alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + alpha_2*alpha_3*(alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + (alpha_4*(alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - alpha_3*(alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*alpha_2*(alpha_3 - alpha_4)*(alpha_1 - alpha_2))*alpha_1)/((alpha_3 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*(alpha_3 - alpha_4)*(alpha_1 - alpha_2))*q_phi/(A*omega**2*rho - k) + Ib*E*(alpha_1*(alpha_3 - alpha_4)*(alpha_2 - alpha_4)*(alpha_2 - alpha_3)*np.exp(-1j*alpha_1*L) - alpha_2*(alpha_3 - alpha_4)*(alpha_1 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*alpha_2*L) + (alpha_1 - alpha_2)*(alpha_3*(alpha_2 - alpha_4)*(alpha_1 - alpha_4)*np.exp(-1j*alpha_3*L) - np.exp(-1j*alpha_4*L)*alpha_4*(alpha_2 - alpha_3)*(alpha_1 - alpha_3)))/((alpha_3 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*(alpha_3 - alpha_4)*(alpha_1 - alpha_2))*q_phi/(A*omega**2*rho - k),
                       -1j*Ib*E*(alpha_4*(alpha_2 - alpha_3)*(alpha_1 - alpha_3)*(alpha_1 - alpha_2)*(alpha_1 + alpha_2 + alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_2 + alpha_3)) - alpha_3*(alpha_2 - alpha_4)*(alpha_1 - alpha_4)*(alpha_1 - alpha_2)*(alpha_1 + alpha_2 + alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_2 + alpha_4)) + (alpha_3 - alpha_4)*(alpha_2*(alpha_1 - alpha_4)*(alpha_1 - alpha_3)*(alpha_1 + alpha_3 + alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_3 + alpha_4)) - np.exp(-1j*L*(alpha_2 + alpha_3 + alpha_4))*alpha_1*(alpha_2 - alpha_4)*(alpha_2 - alpha_3)*(alpha_2 + alpha_3 + alpha_4)))/((alpha_3 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*(alpha_3 - alpha_4)*(alpha_1 - alpha_2))*q_z/(A*omega**2*rho - k) + -1j*Ib*E*(alpha_1*alpha_2*(alpha_1 - alpha_2)*(alpha_1 + alpha_2)*(alpha_3 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_2)) - alpha_1*alpha_3*(alpha_1 - alpha_3)*(alpha_1 + alpha_3)*(alpha_2 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_3)) + alpha_1*alpha_4*(alpha_1 - alpha_4)*(alpha_1 + alpha_4)*(alpha_2 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_4)) + alpha_2*alpha_3*(alpha_2 - alpha_3)*(alpha_2 + alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - alpha_4*(alpha_2*(alpha_2 - alpha_4)*(alpha_2 + alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) - np.exp(-1j*L*(alpha_3 + alpha_4))*alpha_3*(alpha_3 - alpha_4)*(alpha_3 + alpha_4)*(alpha_1 - alpha_2)))/((alpha_3 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*(alpha_3 - alpha_4)*(alpha_1 - alpha_2))*q_z/(A*omega**2*rho - k),
                       (alpha_4*(alpha_2 - alpha_3)*(alpha_1 - alpha_3)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2 + alpha_3)) - alpha_3*(alpha_2 - alpha_4)*(alpha_1 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2 + alpha_4)) + (alpha_3 - alpha_4)*(alpha_2*(alpha_1 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3 + alpha_4)) - np.exp(-1j*L*(alpha_2 + alpha_3 + alpha_4))*alpha_1*(alpha_2 - alpha_4)*(alpha_2 - alpha_3)))*Ib*E/((alpha_3 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*(alpha_3 - alpha_4)*(alpha_1 - alpha_2))*q_phi/(A*omega**2*rho - k) + (alpha_1*alpha_2*(alpha_3 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - alpha_1*alpha_3*(alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + alpha_1*alpha_4*(alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + alpha_2*alpha_3*(alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - alpha_4*(alpha_2*(alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) - np.exp(-1j*L*(alpha_3 + alpha_4))*alpha_3*(alpha_3 - alpha_4)*(alpha_1 - alpha_2)))*Ib*E/((alpha_3 - alpha_4)*(alpha_1 - alpha_2)*np.exp(-1j*L*(alpha_1 + alpha_2)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_1 + alpha_3)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_1 + alpha_4)) + (alpha_2 - alpha_3)*(alpha_1 - alpha_4)*np.exp(-1j*L*(alpha_2 + alpha_3)) - (alpha_2 - alpha_4)*(alpha_1 - alpha_3)*np.exp(-1j*L*(alpha_2 + alpha_4)) + np.exp(-1j*L*(alpha_3 + alpha_4))*(alpha_3 - alpha_4)*(alpha_1 - alpha_2))*q_phi/(A*omega**2*rho - k)])

        return el

    def LocalElementDisplacements(self, u_nodes_global, omega, num_points):
        """
        Calculate local displacements w(s) and rotational displacement phi(s).

        Input:
            u_nodes_global: array. The nodes in global coordinates
            omega: array. Range of frequencies of analysis
            num_points: value. Number of points to divide the element in.
        Output:
            w: array. Amplitude of vertical displacement
            phi: array. Amplitude of rotational displacement
        """
        # get local axis to evaluate on
        L = self.L
        x = np.linspace(0.0, L, num_points)

        # determine nodal displacement in local axis
        u_node_local = self.R @ u_nodes_global

        # extract only the needed displacements
        u_node_local = u_node_local(self.dofs)

        # calculate coeficients
        C = self.Coefficients(u_node_local, omega)

        # get displacement
        w = self.displacement(x, omega, C)

        # get rotations
        phi = self.rotation(x, omega, C)

        return [w, phi]

    def Coefficients(self, u_node_local, omega):
        """
        Calculate the coefficients of the general solution, in this case 4.

        Input:
            u_node_local: local degrees of freedom
            omega: array. Range of frequencies of analysis
        Output:
            C: array (4), coefficients of general solution (C1, C2, C3, C4)
        """
        # Read all the variables
        rho = self.rho
        A = self.A
        L = self.L
        k = self.k
        alpha_1, alpha_2, alpha_3, alpha_4 = self.ElementWaveNumbers(omega)

        # get distributed load value
        q = self.q[1]
        # should be like:
        # q_b, q_m = self.

        # calculate the coefficients with A_mat copy-pasted from Maple
        A_mat = np.array([[1, 1, 1, 1],
                          [1j*alpha_1, 1j*alpha_2, 1j*alpha_3, 1j*alpha_4],
                          [np.exp(-1j*alpha_1*L), np.exp(-1j*alpha_2*L),
                           np.exp(-1j*alpha_3*L), np.exp(-1j*alpha_4*L)],
                          [1j*np.exp(-1j*alpha_1*L)*alpha_1,
                           1j*np.exp(-1j*alpha_2*L)*alpha_2,
                           1j*np.exp(-1j*alpha_3*L)*alpha_3,
                           1j*np.exp(-1j*alpha_4*L)*alpha_4]])

        # TODO - check for correctness
        u_load = np.array([-q/(A*omega**2*rho - k),
                           0,
                           -q/(A*omega**2*rho - k),
                           0])
        C = np.linalg.inv(A_mat) @ (u_node_local + u_load)
        # + np.array([1/(E*Ib*beta_b**4),0,1/(E*Ib*beta_b**4),0]) * q)

        return C

    def displacement(self, x, omega, C=None, u_node_local=None):
        """
        Get the transverse displacments of the EB beam over viscoel foundation.

        Input:
            x: array. Points along element
            omega: array. Range of frequencies of analysis
            C: array. Values of coefficients of general solution
            u_node_local: local nodes
        Ouput:
            w: array. Transverse displacements
        Note:
            if C is not given, then calculate it based on u_node_local.
        """
        alpha_1, alpha_2, alpha_3, alpha_4 = self.ElementWaveNumbers(omega)

        # Check if C is input, otherwise, calculate it
        if C is None:
            C = self.Coefficients(u_node_local, omega)

        # displacements
        w = (C[0]*np.exp(-1j*alpha_1*x) +
             C[1]*np.exp(-1j*alpha_2*x) +
             C[2]*np.exp(-1j*alpha_3*x) +
             C[3]*np.exp(-1j*alpha_4*x))

        return w

    def rotation(self, x, omega, C=None, u_node_local=None):
        """
        Get the rotations of the EB beam over viscoelastic foundation.

        Input:
            x: array. Points along element
            omega: array. Range of frequencies of analysis
            C: array. Values of coefficients of general solution
            u_node_local: local nodes
        Ouput:
            phi: array. Rotational displacements
        Note:
            if C is not given, then calculate it based on u_node_local.
        """
        alpha_1, alpha_2, alpha_3, alpha_4 = self.ElementWaveNumbers(omega)

        # check if C is input
        if C is None:
            C = self.Coefficients(u_node_local, omega)

        # Rotations
        phi = (-1j*C[0]*alpha_1*np.exp(-1j*alpha_1*x) -
               1j*C[1]*alpha_2*np.exp(-1j*alpha_2*x) -
               1j*C[2]*alpha_3*np.exp(-1j*alpha_3*x) -
               1j*C[3]*alpha_4*np.exp(-1j*alpha_4*x))

        return -phi



# -*- coding: utf-8 -*-
"""
Created on Thu Feb 29 13:34:56 2024

@author: rensv
"""

# %% Import dependencies
import numpy as np
from collections import defaultdict
from typing import Optional, Dict
from .dofs import DOF
from .dofs import DOFContainer


# %% class definition
class Node:
    """
    A class to represent a node in a structural system.

    Attributes
    ----------
    nn : int
        Class variable to count the number of nodes.
    dof_configurations : dict
        Dictionary containing DOF configurations for different node types.
    x : float
        x-coordinate of the node.
    z : float
        z-coordinate of the node.
    y : float
        y-coordinate of the node, applicable for 3D configurations.
    config : str
        Configuration type of the node (e.g., '2D', '3D', '2D_torsion').
    dofs : dict
        Dictionary of degrees of freedom for the node with DOF names as keys and their statuses as values.
    nodal_forces : list
        List to store forces applied to the node.
    id : int
        Unique identifier for the node.
    connected_elements : list
        List of elements connected to this node.
    """
    
    # initialise number of nodes (nn) as class-variable, to be updated by every new class object
    nn   = 0
    
    # Node configurations are ordered as: 'Name': [['linear dofs'],['rotational dofs']]. Linear dofs also define the amount of coordinates we can assign (thus 2D or 3D basically)
    dof_configurations = {
        '2D': [['x', 'z'],['phi_y']],
        '3D': [['x', 'z', 'y'], ['phi_x', 'phi_z', 'phi_y']],
        '2D_torsion': [['x', 'z'], ['phi_x', 'phi_y']] 
        }
    
    def __init__(self,x,z,y=0,config='2D', dof_config=None):
        """
        Initializes the Node with its coordinates and DOF configuration.

        Parameters
        ----------
        x : float
            x-coordinate of the node.
        z : float
            z-coordinate of the node.
        y : float, optional
            y-coordinate of the node, default is 0 (for 2D configurations).
        config : str, optional
            Configuration type of the node, default is '2D'.
        """
        
        # dofs of the node: {dof: value} - None = Free, 0 = Fixed, value = presribed
        self.config = config
        
        # Load any default dof_config or apply custom one SHOULD THINK WHETHER THIS IS USEFULL OR JUST LEAVE IT FULLY AUTOMATIC BASED ON APPLIED SECTIONS
        self.dof_config = dof_config if dof_config else self.dof_configurations[config]
        
        # initialise the DOFcontainer
        self.dof_container = DOFContainer()
        
        # Set up DOFs based on configuration
        for dof_list in self.dof_config:
            for dof_name in dof_list:
                self.dof_container.set_dof(dof_name)
                
        # set the location in space
        self.x     = x
        self.z     = z
        self.y     = y
        
        # node name
        self.id = Node.nn
        Node.nn   += 1 # increment the class variables
        
        # Initialise empty lists needed for later
        self.nodal_loads = defaultdict(dict)
        self.connected_elements = []
        
    def get_dof_indices(self):
        """Returns a list of global DOF indices for the node."""
        return [dof.index for dof in self.dof_container.dofs.values() if dof.index is not None]

    def update_element_dof(self,changes=None):
        """
        Updates the DOFs of connected elements based on changes in the node's DOFs.

        Parameters
        ----------
        changes : dict, optional
            Dictionary of changes in DOFs, default is None.
        """
        if changes:
            for element in self.connected_elements:
                element.update_node_dofs(self,changes)
                
    def apply_dof_change_to_elements(self, dof_name, value):
        """
        Applies the change in a specific DOF value to the connected elements,
        mapping the global DOF to the corresponding local and global DOFs in each element.
    
        Parameters
        ----------
        dof_name : str
            The name of the global DOF that has changed (e.g., 'x', 'phi_y').
        value : Any
            The value to set for the specified DOF.
        """
        # Get the DOF from the container
        dof = self.dof_container.get_dof(dof_name)
        
        # Ensure the DOF exists and is part of the node's current configuration
        if dof is None:
            raise ValueError(f"DOF '{dof_name}' is not part of the node's current configuration.")
        
        # Propagate the change to all connected elements if present
        if not self.connected_elements:
            return
        else:
            for element in self.connected_elements:
                element.apply_global_dof_change(self, dof_name, value)
    
        print(f"Global DOF '{dof_name}' with value '{value}' applied to connected elements.")               

    def connect_element(self, element):
        """
        Connects an element to this node.

        Parameters
        ----------
        element : Element
            The element to be connected to this node.
        """
        if element not in self.connected_elements:
            self.connected_elements.append(element)
            
    def fix_node(self, *dofs):
        """
        Fixes specified DOFs of the node to zero.

        Parameters
        ----------
        *dofs : str
            DOFs to be fixed.
        """
    
        self.prescribe_node(**{dof: 0 for dof in dofs})
        
    def free_node(self, *dofs):
        """
        Frees specified DOFs of the node (sets them to None).

        Parameters
        ----------
        *dofs : str
            DOFs to be freed.
        """
        
        self.prescribe_node(**{dof: None for dof in dofs})
       
    def prescribe_node(self, **dofs):
        """
        Sets specified DOFs of the node to given values.

        Parameters
        ----------
        **dofs : dict
            DOFs to be prescribed with their values.
        
        Raises
        ------
        ValueError
            If any of the specified DOFs are not in the current configuration.
        """

        try:
            for dof_name, value in dofs.items():
                if not self.dof_container.has_dof(dof_name):
                    raise ValueError(f"DOF '{dof_name}' is not available in the current configuration. Available DOFs: {self.dof_config}")
                
                dof = self.dof_container.get_dof(dof_name)
                if dof.value != value:
                    # set value if is not the same
                    dof.value = value
                    # also apply the change to the connected elements
                    self.apply_dof_change_to_elements(dof_name,value)
        except ValueError as e:
            print(e)
            return
                
    def add_load(self,**loads):
        '''
        Adds distributed loads to the Node.
    
        Parameters
        ----------
        **loads : dict
            Keyword arguments where the key is the DOF (degree of freedom) and the value is the load magnitude. e.g. node.AddLoad(x=load_x, y=load_y
        '''
        
        # Assign unpacked loads to the element's list of loads
        for dof, load in loads.items():
            # check if dof is present
            if self.dof_container.has_dof(dof):
                if dof not in self.nodal_loads.keys():
                    self.nodal_loads[dof] = load
                else:
                    print(f'Load already added for DOF {dof}')  
            else:
                print(f'Node {self.node_id} does not contain DOF {dof}. Load has not been added.')                    

    def get_coords(self):
        """
        Returns the coordinates of the node based on its DOF configuration.

        Returns
        -------
        tuple
            Coordinates of the node (x, z, y) if applicable.
        """
        coords = [self.x, self.z, self.y]

        return tuple(coords)



# -*- coding: utf-8 -*-
"""
Created on Wed Jun 12 11:52:01 2024

@author: rensv
"""

# Import dependencies
import numpy as np
from scipy.linalg import inv
from ..elements import ElementFactory
from collections import defaultdict

# Class definition
class Analysis:
    """
    A class to handle structural analysis computations such as stiffness matrices, force vectors, and displacement calculations.
    """

    def GlobalStiffness(self, nodes, elements, omega):
        """
        Assembles the global unconstrained stiffness matrix.

        Parameters
        ----------
        nodes : list of Node
            List of nodes in the structural system.
        elements : list of Element
            List of elements in the structural system.
        omega : float
            Frequency parameter.

        Returns
        -------
        k_global : numpy.ndarray
            Global stiffness matrix.
        """
        # load k_global
        k_global = np.zeros((self.num_dofs, self.num_dofs), complex)
        
        # set up block matrix per element
        for e in elements:
            elmat = e.Stiffness(omega)
            idofs = e.get_element_node_dof_indices_global()
            k_global[np.ix_(idofs, idofs)] = elmat
            
        # sort by unique / redundant dofs to comply with the structure of L
        k_uu = k_global[np.ix_(self.unique_dofs,self.unique_dofs)]
        k_ur = k_global[np.ix_(self.unique_dofs,self.redundant_dofs)]
        k_ru = k_global[np.ix_(self.redundant_dofs,self.unique_dofs)]
        k_rr = k_global[np.ix_(self.redundant_dofs,self.redundant_dofs)]
        
        k_global = np.block([[k_uu, k_ur],
                                    [k_ru, k_rr]])
        
        # constrain by using L
        K_global = self.L.T @ k_global @ self.L
    
        return K_global

    def GlobalForce(self, nodes, elements, omega):
        """
        Assembles the global unconstrained force vector.

        Parameters
        ----------
        nodes : list of Node
            List of nodes in the structural system.
        elements : list of Element
            List of elements in the structural system.
        omega : float
            Frequency parameter.

        Returns
        -------
        f_global : numpy.ndarray
            Global force vector.
        """
        
        # set the global force vector
        f_global = np.zeros(self.num_dofs, complex)
        
        # get all element loads
        for element in elements:
            # if no loads present continue loop
            if not element.element_loads:
                continue
            
            # get dofs of the element
            dofs = element.get_element_node_dof_indices_global()
            
            # for element_load in element.element_loads:
            f_global[np.ix_(dofs)] += element.EvaluateDistributedLoad(element.element_loads, omega)
        
        # get all nodal loads - ASSUMING EVERYTHING IS ALREADY WITHIN GLOBAL DIRECTIONS 
        for node in nodes:
            if not node.nodal_loads:
                continue
            
            # load the dof container
            node_dof_container = node.dof_container
            
            # loop over nodal loads
            for dof_name, load in node.nodal_loads.items():
                # add the load to the dof if dof is present in the current dof config of that node (probably should add this check in the nodal class..)
                if node_dof_container.has_dof(dof_name):
                    dof = node_dof_container.dofs[dof_name]
                    index = dof.index
                    if index is not None:
                        value = load(omega) if callable(load) else load
                        f_global[index] += value
        
        # sort by unique / redundant dofs to comply with the structure of L
        f_global_u = f_global[np.ix_(self.unique_dofs)]
        f_global_r = f_global[np.ix_(self.redundant_dofs)]
        
        f_global = np.hstack((f_global_u,f_global_r))
        
        # constrain by using L
        F_global = self.L.T @ f_global 
        
        return F_global

    def GlobalConstrainedStiffness(self, nodes, elements, omega):
        """
        Constrains the global stiffness matrix.

        Parameters
        ----------
        nodes : list of Node
            List of nodes in the structural system.
        elements : list of Element
            List of elements in the structural system.
        omega : float
            Frequency parameter.

        Returns
        -------
        k_constrained : numpy.ndarray
            Constrained global stiffness matrix.
        """
        k_global = self.GlobalStiffness(nodes, elements, omega)
        return k_global[np.ix_(self.free_dofs, self.free_dofs)]

    def GlobalConstrainedForce(self, nodes, elements, omega):
        """
        Constrains the global force vector.

        Parameters
        ----------
        nodes : list of Node
            List of nodes in the structural system.
        elements : list of Element
            List of elements in the structural system.
        omega : float
            Frequency parameter.

        Returns
        -------
        f_constrained : numpy.ndarray
            Constrained global force vector.
        """
        k_global = self.GlobalStiffness(nodes, elements, omega)
        f_global = self.GlobalForce(nodes, elements, omega)
        
        constrained_indices = list(self.constrained_dofs.keys())
        constrained_values = list(self.constrained_dofs.values())
        
        K_free_constrained = k_global[np.ix_(self.free_dofs, constrained_indices)]
        F_free = f_global[self.free_dofs]
        
        return F_free - K_free_constrained @ constrained_values

    def SolveUfree(self, Kc_global, fc_global):
        """
        Solves the free displacements.

        Parameters
        ----------
        Kc_global : numpy.ndarray
            Constrained global stiffness matrix.
        fc_global : numpy.ndarray
            Constrained global force vector.

        Returns
        -------
        u_free : numpy.ndarray
            Free displacements.
        """
        return np.linalg.inv(Kc_global) @ fc_global

    def SupportReactions(self, k_global, u_free, f_global):
        """
        Calculates the support reactions.

        Parameters
        ----------
        k_global : numpy.ndarray
            Global stiffness matrix.
        u_free : numpy.ndarray
            Free displacements.
        f_global : numpy.ndarray
            Global force vector.

        Returns
        -------
        reactions : numpy.ndarray
            Support reactions.
        """
        constrained_indices = list(self.constrained_dofs.keys())
        constrained_values = list(self.constrained_dofs.values())
        
        Kcf = k_global[np.ix_(constrained_indices, self.free_dofs)]
        Kcc = k_global[np.ix_(constrained_indices, constrained_indices)]
        
        return (Kcf @ u_free) + (Kcc @ constrained_values) - f_global[np.ix_(constrained_indices)]

    def FullDisplacement(self, u_free):
        """
        Returns the full displacement vector for the entire structure.

        Parameters
        ----------
        u_free : numpy.ndarray
            Free displacements.

        Returns
        -------
        u_full : numpy.ndarray
            Full displacement vector.
        """
        constrained_indices = list(self.constrained_dofs.keys())
        constrained_values = list(self.constrained_dofs.values())
        
        u_full = np.zeros(len(self.free_dofs) + len(constrained_indices), dtype=complex)
        
        u_full[np.ix_(self.free_dofs)] = u_free
        u_full[np.ix_(constrained_indices)] = constrained_values
        
        return u_full

    def SolveEigenvector(self, nodes, elements, omega, fixed_index=0):
            """
            Computes an eigenvector for the system using the constrained global stiffness matrix.
    
            Parameters
            ----------
            nodes : list of Node
                List of nodes in the structural system.
            elements : list of Element
                List of elements in the structural system.
            omega : float
                Frequency parameter.
            fixed_index : int, optional
                Index of the eigenvector component to be set to 1 (default is 0).
    
            Returns
            -------
            U : numpy.ndarray
                Computed eigenvector normalized by fixing one component to 1.
            """
            # Retrieve the constrained global stiffness matrix
            K_constrained = self.GlobalConstrainedStiffness(nodes, elements, omega)
    
            # Ensure the matrix is not singular
            if np.linalg.matrix_rank(K_constrained) < K_constrained.shape[0]:
                raise ValueError("Singular stiffness matrix, the system may be improperly constrained.")
    
            # Define the number of DOFs
            num_dofs = K_constrained.shape[0]
    
            # Partition the matrix explicitly
            free_dofs = [i for i in range(num_dofs) if i != fixed_index]
    
            # Extract submatrices
            K_rr = K_constrained[np.ix_(free_dofs, free_dofs)]  # Reduced stiffness matrix
            K_rp = K_constrained[np.ix_(free_dofs, [fixed_index])]  # Column corresponding to fixed index
    
            # Solve for the unknown components
            try:
                U_r = -np.linalg.solve(K_rr, K_rp).flatten()  # Solve for U_r
            except np.linalg.LinAlgError:
                raise ValueError("Numerical issue: unable to solve for eigenvector.")
                
            # Reconstruct the full eigenvector
            U = np.zeros(num_dofs, dtype=complex)
            U[fixed_index] = 1  # Set fixed value
            U[free_dofs] = U_r  # Assign solved components
    
            return U   

    def ElementDisplacements(self, elements, u_nodes_global, omega, num_points=20):
        """
        Gathers the displacements for each element from the element-level calculations.
    
        Parameters
        ----------
        elements : list of Element
            List of elements in the structural system.
        u_nodes_global : numpy.ndarray
            Global displacements of nodes.
        omega : float
            Frequency parameter.
        num_points : int, optional
            Number of points along the element to calculate displacements, default is 20.
    
        Returns
        -------
        element_displacements : dict
            Dictionary with element IDs as keys and their global displacements (as numpy arrays) as values.
        """
        
        # Apply L matrix to get displacements of redundant nodes as well
        u_nodes_global_all = self.L @ u_nodes_global
        # Get the current order
        current_order = self.unique_dofs + self.redundant_dofs
        
        # Create an index map to reorder the elements to match sorted_order
        sorted_order = sorted(current_order)
        index_map = [current_order.index(idx) for idx in sorted_order]
        
        # Reorder u_nodes_global_all using the index map
        u_nodes_global_all_sorted = np.array(u_nodes_global_all)[index_map]
        
        element_displacements = {}
    
        for element in elements:
            # Retrieve global nodal DOF indices for the current element
            global_dof_indices = element.get_node_dof_indices_global()
    
            # Extract the relevant global displacements for this element
            u_element_global = u_nodes_global_all_sorted[global_dof_indices]
    
            # Calculate the displacements for the element using its Displacements method
            u_elem = element.Displacements(u_element_global, omega, num_points)
    
            # Store the displacements in the dictionary using the element's ID
            element_displacements[element.id] = u_elem
    
        return element_displacements
    
    def ElementForces(self, elements, u_nodes_global, omega, num_points=20):
        """
        Gathers the displacements for each element from the element-level calculations.
    
        Parameters
        ----------
        elements : list of Element
            List of elements in the structural system.
        u_nodes_global : numpy.ndarray
            Global displacements of nodes.
        omega : float
            Frequency parameter.
        num_points : int, optional
            Number of points along the element to calculate displacements, default is 20.
    
        Returns
        -------
        element_forces : dict
            Dictionary with element IDs as keys and their global displacements (as numpy arrays) as values.
        """
        
        # Apply L matrix to get displacements of redundant nodes as well
        u_nodes_global_all = self.L @ u_nodes_global
        # Get the current order
        current_order = self.unique_dofs + self.redundant_dofs
        
        # Create an index map to reorder the elements to match sorted_order
        sorted_order = sorted(current_order)
        index_map = [current_order.index(idx) for idx in sorted_order]
        
        # Reorder u_nodes_global_all using the index map
        u_nodes_global_all_sorted = np.array(u_nodes_global_all)[index_map]
        
        element_forces = {}
    
        for element in elements:
            # Retrieve global nodal DOF indices for the current element
            global_dof_indices = element.get_node_dof_indices_global()
    
            # Extract the relevant global displacements for this element
            u_element_global = u_nodes_global_all_sorted[global_dof_indices]
    
            # Calculate the displacements for the element using its Displacements method
            f_elem = element.Forces(u_element_global, omega, num_points)
    
            # Store the displacements in the dictionary using the element's ID
            element_forces[element.id] = f_elem
    
        return element_forces

    def ElementStresses(self, elements, u_nodes_global, omega, num_points=20):
        """
        Gathers the stresses for each element from the element-level calculations.
    
        Parameters
        ----------
        elements : list of Element
            List of elements in the structural system.
        u_nodes_global : numpy.ndarray
            Global displacements of nodes.
        omega : float
            Frequency parameter.
        num_points : int, optional
            Number of points along the element to calculate displacements, default is 20.
    
        Returns
        -------
        element_stresses : dict
            Dictionary with element IDs as keys and their global displacements (as numpy arrays) as values.
        """
        
        # Apply L matrix to get displacements of redundant nodes as well
        u_nodes_global_all = self.L @ u_nodes_global
        # Get the current order
        current_order = self.unique_dofs + self.redundant_dofs
        
        # Create an index map to reorder the elements to match sorted_order
        sorted_order = sorted(current_order)
        index_map = [current_order.index(idx) for idx in sorted_order]
        
        # Reorder u_nodes_global_all using the index map
        u_nodes_global_all_sorted = np.array(u_nodes_global_all)[index_map]
        
        element_stresses = {}
    
        for element in elements:
            # Retrieve global nodal DOF indices for the current element
            global_dof_indices = element.get_node_dof_indices_global()
    
            # Extract the relevant global displacements for this element
            u_element_global = u_nodes_global_all_sorted[global_dof_indices]
    
            # Calculate the displacements for the element using its Displacements method
            s_elem = element.Stresses(u_element_global, omega, num_points)
    
            # Store the displacements in the dictionary using the element's ID
            element_stresses[element.id] = s_elem
    
        return element_stresses
    
# %% specific methods for this method of analysis

    def find_unique_redundant_dofs(self, B):
        """
        Identifies unique and redundant degrees of freedom (DOFs) based on the constraint matrix B.

        Parameters
        ----------
        B : numpy.ndarray
            The constraint matrix where rows represent constraints and columns represent DOFs.

        Returns
        -------
        unique_dofs : list of int
            Sorted list of unique DOFs.
        redundant_dofs : list of int
            Sorted list of redundant DOFs.
        """
        self.num_dofs = B.shape[1]
        redundant_dofs = set()
        unique_dofs = set()

        for row in B:
            positives = np.where(row > 0)[0]
            negatives = np.where(row < 0)[0]

            for pos in positives:
                unique_dofs.add(pos)
            for neg in negatives:
                redundant_dofs.add(neg)

        # Only columns that are completely zero and not identified in negatives are truly unique
        zero_dofs = {i for i in range(self.num_dofs ) if np.all(B[:, i] == 0)}
        unique_dofs = (unique_dofs | zero_dofs) - redundant_dofs

        return sorted(unique_dofs), sorted(redundant_dofs)

    def calculate_L(self, B, unique_dofs, redundant_dofs):
        """
        Calculates the transformation matrix L based on the constraint matrix B.

        Parameters
        ----------
        B : numpy.ndarray
            The constraint matrix where rows represent constraints and columns represent DOFs.
        unique_dofs : list of int
            List of unique DOFs.
        redundant_dofs : list of int
            List of redundant DOFs.

        Returns
        -------
        L : numpy.ndarray
            The transformation matrix L.

        Raises
        ------
        ValueError
            If the matrix B_r is singular and cannot be inverted.
        """
        # check if there are any redundant dofs
        if len(redundant_dofs) != 0:
            B_r = B[:, redundant_dofs]
            B_u = B[:, unique_dofs]
    
            try:
                B_r_inv = inv(B_r)
                L_lower = -B_r_inv @ B_u
                L = np.vstack((np.eye(len(unique_dofs)), L_lower))
                return L
            
            except np.linalg.LinAlgError:
                raise ValueError("Matrix B_r is singular and cannot be inverted.")
        # if not, return L solely based on unique dofs
        else:
            L = np.eye(len(unique_dofs))
            return L

    def assign_dof_indices_old(self, nodes, elements):
        """
        Assigns global indices to the degrees of freedom (DOFs) of all elements.
        The element dofs are leading here, nodal dofs will be used to connect element dofs

        Parameters
        ----------
        elements : list of Element
            List of elements in the structural system.

        Returns
        -------
        dof_indices : dict
            Dictionary mapping (node_id, element_id) to a dict of DOFs and their global indices.
        """
        # TODO - currently we loop first over nodes which felt that it makes sense. However the global matrix before constraining by L is broken up. There are no block matrices per element in that case
        dof_indices = defaultdict(dict)
        global_index = 0  # Start a global index counter for all DOFs in the system

        for node in nodes:
            for element in node.connected_elements:
                
                # use the DOF configuration of the element for this specific node
                for dof in element.dofs[node.id].keys():
                    dof_indices[(node.id, element.id)][dof] = global_index
                    # also assign to the element itself as this can be useful
                    element.dof_indices[node.id][dof] = global_index
                    global_index += 1  # Increment global index for each DOF

        return dof_indices
    
    def assign_dof_indices(self, nodes, elements):
        """
        Assigns global indices to the degrees of freedom (DOFs) of all elements.
        The element DOFs are leading here; nodal DOFs will be used to connect element DOFs.
    
        Parameters
        ----------
        nodes : list of Node
            List of nodes in the structural system.
        elements : list of Element
            List of elements in the structural system.
    
        Returns
        -------
        dof_indices : defaultdict
            Dictionary mapping (node_id, element_id) to a dict of DOFs and their global indices.
        """
        dof_indices = defaultdict(dict)
        global_index = 0  # Start a global index counter for all DOFs in the system
        
        for element in elements:
    
            for node in element.nodes:
                node_id = node.id
                element_dof_container = element.dof_containers[node.id]
                node_dof_container = node.dof_container
                
                # Use the DOF configuration of the element for this specific node
                for dof_name, dof in element_dof_container.dofs.items():
                    dof_indices[(node_id, element.id)][dof_name] = global_index
                    
                    # Assign the index also to the dof in the global dof container of the element
                    dof.index = global_index
                    
                    # TODO - SEE HOW TO HANDLE DOF INDICES FOR ELEMENT DOFS AND NODE DOFS IN A NEAT WAY. NOW WE HAVE THREE DIFFERENT DICTIONARIES STORING THEM. NOT THAT EFFICIENT ATM
                    # same for the node, but only if present in node
                    if node_dof_container.has_dof(dof_name):
                        # add index only if it hasn't been assigned yet - # TODO this may not be that useful..
                        if not node_dof_container.dofs[dof_name].index:
                            node_dof_container.dofs[dof_name].index = global_index
                        
                    global_index += 1  # Increment global index for each DOF
        
        num_dof = global_index                    
        return dof_indices, num_dof


    def build_matrix_B(self, nodes, elements, dof_indices, num_dof):
        """
        Builds the constraint matrix B for the structural system.
    
        Parameters
        ----------
        nodes : list of Node
            List of nodes in the structural system.
        elements : list of Element
            List of elements in the structural system.
        dof_indices : dict
            Dictionary mapping (node_id, element_id) to a dict of DOFs and their global indices.
    
        Returns
        -------
        B : numpy.ndarray
            The constraint matrix B.
        """
        # num_dof = sum(len(dofs) for element in elements for dofs in element.dof_container.values())  # Calculate total DOFs based on the elements global dofs at each connection
    
        # Preallocate a larger matrix to avoid expensive np.vstack() operations
        B = np.zeros((num_dof, num_dof))
        constraint_counter = 0  # Keeps track of which row in B we are filling
    
        for node in nodes:
            if len(node.connected_elements) > 1:  # Only consider nodes connected to multiple elements
                # find the element classes and the indices of its dofs that are connected to the current node
                connected_indices = [(element, dof_indices[(node.id, element.id)])  
                                     for element in node.connected_elements
                                     if (node.id, element.id) in dof_indices]
                
                # only continue if connected indices are actually found
                if connected_indices:
                    # 
                    for dof_name, nodal_dof in node.dof_container.dofs.items():
                        # Collect DOFs from elements that are monolithically connected
                        all_dofs = [
                            indices[dof_name] for element, indices in connected_indices
                            if dof_name in indices and element.dof_containers[node.id].dofs[dof_name].value == nodal_dof.value
                            and element.constraint_types[node.id][dof_name] == "monolithic"
                        ]
                        # continue only if we still have connected dofs
                        if len(all_dofs) > 1:
                            # Use the first DOF as the reference for compatibility
                            primary_index = all_dofs[0]
                            for index in all_dofs[1:]:
                                row = np.zeros(num_dof)
                                row[primary_index] = 1  # Primary DOF
                                row[index] = -1  # Enforce compatibility with other DOFs
                                B[constraint_counter, :] = row  # Fill row directly instead of stacking
                                constraint_counter += 1
    
        # Return the trimmed matrix if we overestimated the number of constraints
        return B[:constraint_counter, :]


    def connectivity(self, nodes, elements):
        """
        Computes the connectivity information for the structural system.

        Parameters
        ----------
        nodes : list of Node
            List of nodes in the structural system.
        elements : list of Element
            List of elements in the structural system.

        Returns
        -------
        B : numpy.ndarray
            The constraint matrix B.
        L : numpy.ndarray
            The transformation matrix L.
        dof_indices : dict
            Dictionary mapping (node_id, element_id) to a dict of DOFs and their global indices.
        unique_dofs : list of int
            Sorted list of unique DOFs.
        redundant_dofs : list of int
            Sorted list of redundant DOFs.
        free_dofs: list
            List of indices in `unique_dofs` that are free.
        constrained_dofs: list
            List of lists where each sublist contains a DOF index and its value.
        """
        
        self.dof_indices, num_dof = self.assign_dof_indices(nodes, elements)
        
        self.B = self.build_matrix_B(nodes, elements, self.dof_indices,num_dof)
        
        self.unique_dofs, self.redundant_dofs = self.find_unique_redundant_dofs(self.B)
        
        self.L = self.calculate_L(self.B, self.unique_dofs, self.redundant_dofs)
        
        self.free_dofs, self.constrained_dofs = self.classify_free_constrained_dofs(nodes, elements, np.array(self.unique_dofs), self.dof_indices)

    
    def classify_free_constrained_dofs_old(self, nodes, elements, unique_dofs, dof_indices):
        """
        Classifies unique DOFs as free or constrained (fixed or prescribed) and extracts their indices.
        
        Parameters
        --------
        nodes: list of Node
            List of all nodes.
        elements: list of Element
            List of elements in the structural system.
        unique_dofs: list
            List of indices considered unique.
        dof_indices: dict
            Maps (node_id, element_id) to a dict of DOFs and their indices.
        
        Returns
        -------
        free_dofs: list
            List of indices in `unique_dofs` that are free.
        constrained_dofs: list
            List of lists where each sublist contains a DOF index and its value.
        """
        
        # create dict to look up via O(1) instead of O(n) for larger system        
        unique_dof_dict = {old_index: idx for idx, old_index in enumerate(unique_dofs)}

        # introduce empty lists
        free_dofs = []
        constrained_dofs = {}
        
        # Iterate over all nodes and their connected elements to classify DOFs
        for node in nodes:
            
            # loop over all connected elements
            for element in node.connected_elements:
                # get element dofs
                element_dofs = element.dofs[node.id]
                # get dofs from dof_indices dictionary
                for dof, old_index in dof_indices[(node.id,element.id)].items():
                    
                    # check whether the current dof is in the unique dofs
                    new_index = unique_dof_dict.get(old_index)
                    
                    # only continue if only 1 location is found
                    if new_index is not None:
                        
                        # get value from element
                        dof_value = element_dofs[dof]
                        
                        # set accordingly, i.e. if None its a free node otherwise its prescribed
                        if dof_value is None:
                            free_dofs.append(new_index)
                        else:
                            constrained_dofs[new_index] = dof_value
        
         # TODO - return sorted list, not per se needed I guess. Furthermore I could change the constrained dofs list to a dictionary
        # return sorted(free_dofs), sorted(constrained_dofs, key=lambda x: x[0])
        return np.array(free_dofs), constrained_dofs
    
    def classify_free_constrained_dofs(self, nodes, elements, unique_dofs, dof_indices):
        """
        Classifies unique DOFs as free or constrained (fixed or prescribed) and extracts their indices.
    
        Parameters
        ----------
        nodes : list of Node
            List of all nodes.
        elements : list of Element
            List of elements in the structural system.
        unique_dofs : list
            List of indices considered unique.
        dof_indices : dict
            Maps (node_id, element_id) to a dict of DOFs and their indices.
    
        Returns
        -------
        free_dofs : numpy.ndarray
            Array of indices in `unique_dofs` that are free.
        constrained_dofs : dict
            Dictionary mapping DOF indices to their prescribed values.
        """
        # Create a mapping from old DOF indices to new indices in unique_dofs
        unique_dof_dict = {old_index: idx for idx, old_index in enumerate(unique_dofs)}
    
        free_dofs = []
        constrained_dofs = {}
    
        # Iterate over all elements and their nodes to classify DOFs
        for element in elements:
            for node in element.nodes:
                node_id = node.id
                element_id = element.id
                element_dof_container = element.dof_containers[node_id]
    
                # Get DOFs from dof_indices dictionary
                for dof_name, old_index in dof_indices[(node_id, element_id)].items():
                    # Check whether the current DOF is in the unique DOFs
                    new_index = unique_dof_dict.get(old_index)
    
                    if new_index is not None:
                        # Get value from element's DOFContainer
                        dof_value = element_dof_container.get_dof(dof_name).value
    
                        # Classify DOF as free or constrained
                        if dof_value is None:
                            free_dofs.append(new_index)
                        else:
                            constrained_dofs[new_index] = dof_value
    
        return np.array(free_dofs), constrained_dofs


# %% import dependencies

import numpy as np
from collections import defaultdict
from ..elements import ElementFactory
from .dofs import DOFContainer, DOF




# %% class definition
class Element:
    """
    A class to represent an element in a structural system.

    Attributes
    ----------
    ne : int
        Class variable to count the number of elements.
    id : int
        Unique identifier for the element.
    nodes : list of Node
        List of nodes connected to this element.
    dofs : dict
        Dictionary of DOFs for each node in the element.
    """
    
    # initialise number of elements (ne) as class-variable, to be updated by every new class object
    ne = 0
    
    # dof mapping in full space and its reverse
    dof_mapping = {
        'x': 0,
        'z': 1,
        'y': 2,
        'phi_x': 3,
        'phi_z': 4,
        'phi_y': 5
    }
    
    # Ndof is essentially the length of dof_mapping, which is 6
    maxNdof = len(dof_mapping)
    
    reverse_dof_mapping = {v: k for k, v in dof_mapping.items()}
    
    def __init__(self, nodes):
        """
        Initializes the Element with its connected nodes.

        Parameters
        ----------
        nodes : list of Node
            List of nodes connected to this element.
        """
        
        # Element number
        self.id = Element.ne
        Element.ne += 1 # increment element number
        
        # reference the nodes
        self.nodes = nodes
        
        # connect the element to the node
        for node in nodes:
            node.connect_element(self)  # Connect this element to its nodes
                
        # Initialize DOFContainers for global DOFs, 
        self.dof_containers = {}
        
        # initalise constraint type dict - options: monolithic - independent 
        self.constraint_types = {node.id: {} for node in nodes}
        
        # copying DOFs from nodes
        for node in nodes:
            # Load nodal DOFContainer
            node_dof_container = node.dof_container
            # intialise element DOFContainer
            element_dof_container = DOFContainer()
            
            # Copy DOFs from node's DOFContainer to element's DOFContainer
            for dof_name, node_dof in node_dof_container.dofs.items():
                # Create a new DOF instance with the same name and value
                element_dof = DOF(name=dof_name, value=node_dof.value, index=node_dof.index)
                element_dof_container.dofs[dof_name] = element_dof
                
                # set dof to monolithic initially
                self.constraint_types[node.id][dof_name] = 'monolithic'
                
            self.dof_containers[node.id] = element_dof_container
        
        # set local dofs containers as empty first
        self.local_dof_container = {node.id: DOFContainer() for node in nodes}        
        
        # calculate geometrical properties
        self.geometrical_properties()
        
        # Keep track of what kind of elements are included in the form of a dictionary to be able to use them later
        self.element_types = {}
        
        # initialise empty list with nodal forces due to element loads
        self.element_loads = defaultdict(dict)

# %% Nodes and Dofs

    def get_element_node_dof_indices_global(self):
        return [dof.index for node in self.nodes for dof in self.dof_containers[node.id].dofs.values()]        
    
    def get_node_dof_indices_global(self):
        return [dof_index for node in self.nodes for dof_index in node.get_dof_indices()]   

    def check_dofs(self, dofs):
        """
        Checks which DOFs are present and returns their numerical values.

        Parameters
        ----------
        dofs : list of str
            List of DOFs to check (e.g., ['x', 'z']).

        Returns
        -------
        numerical_dofs : list of int
            List of numerical values corresponding to the DOFs.
        """
        numerical_dofs = []
        for dof_name in dofs:
            if dof_name in self.dof_mapping:
                numerical_dofs.append(self.dof_mapping[dof_name])
            else:
                raise ValueError(f"DOF '{dof_name}' is not recognized.")
        return numerical_dofs

    def get_element_type_dofs(self, element_type):
        '''
        Translates the dofs of an element to the nodal dofs;
        
        Example:
            Rod element has dof = 0 (axial only)
            This will give: nodal_dofs = [0,3]
        '''
        
        element_dofs = self.check_dofs(element_type.dofs)
        
        maxNdof = Element.maxNdof
        
        element_type_dofs = []
        
        # Process each DOF to map it to the nodes' corresponding DOFs
        for i in range(2):
            for dof in element_dofs:
                # Calculate the DOF index for the node
                node_dof_index = dof + i * maxNdof
                element_type_dofs.append(node_dof_index)
                
        return element_type_dofs
    
    
    def get_full_element_dof_indices_global(self) -> list:
        """
        Returns the list of global DOF indices for the element based on the DOFs stored in element.dof_containers
        Based on the numbering provided by class variable 'dof_mapping'
    
        Returns
        -------
        nodal_dofs : list of int
            List of global DOF indices for the element's DOFs.
        """
        element_dof_indices = []
        Ndof = Element.maxNdof  
        
        for i, node in enumerate(self.nodes):
            element_dof_container = self.dof_containers[node.id]
            for dof_name, dof in element_dof_container.dofs.items():
                # Calculate local index if global index is not assigned
                local_dof_index = self.dof_mapping[dof_name] + i * Ndof
                element_dof_indices.append(local_dof_index)
        
        return element_dof_indices
    
    def get_full_element_dof_indices_local(self) -> list:
        """
        Returns the list of global DOF indices for the element based on the DOFs stored in element.local_dof_container
        Based on the numbering provided by class variable 'dof_mapping'
    
        Returns
        -------
        nodal_dofs : list of int
            List of local DOF indices for the element's DOFs.
        """
        element_dof_indices = []
        Ndof = Element.maxNdof  
        
        for i, node in enumerate(self.nodes):
            element_dof_container = self.local_dof_container[node.id]
            for dof_name, dof in element_dof_container.dofs.items():
                # Calculate local index if global index is not assigned
                local_dof_index = self.dof_mapping[dof_name] + i * Ndof
                element_dof_indices.append(local_dof_index)
        
        return element_dof_indices
    
    def get_specific_dof_indices_global(self, dofs) -> list:
        """
        Returns the list of indices for the specified DOFs that are actually present within the element.
    
        This method returns the indices for the specified DOFs that are defined for each node within
        the element. For example, if one node has DOFs 'x' and 'z', and another has only 'x', and you
        are querying for 'x', the returned indices will be those corresponding to 'x' for both nodes.
    
        Parameters
        ----------
        dofs : list of str
            List of DOFs to find indices for (e.g., ['x', 'z']).
    
        Returns
        -------
        specific_dof_indices : list of int
            List of indices representing the specified DOFs that are currently defined in the element.
        """
        specific_dof_indices = []
        index_counter = 0
    
        for node in self.nodes:
            element_dof_container = self.dof_containers[node.id]  
            for dof_name in element_dof_container.dofs.keys():
                # Check if the current DOF is in the list of required DOFs
                if dof_name in dofs:
                    specific_dof_indices.append(index_counter)
                index_counter += 1
    
        return specific_dof_indices
    
    def get_specific_dof_indices_local(self, dofs) -> list:
        """
        Returns the list of indices for the specified DOFs that are actually present within the element.
    
        This method returns the indices for the specified DOFs that are defined for each node within
        the element. For example, if one node has DOFs 'x' and 'z', and another has only 'x', and you
        are querying for 'x', the returned indices will be those corresponding to 'x' for both nodes.
    
        Parameters
        ----------
        dofs : list of str
            List of DOFs to find indices for (e.g., ['x', 'z']).
    
        Returns
        -------
        specific_dof_indices : list of int
            List of indices representing the specified DOFs that are currently defined in the element.
        """
        specific_dof_indices = []
        index_counter = 0
    
        for node in self.nodes:
            element_dof_container = self.local_dof_container[node.id]  # Use local DOF container
            for dof_name in element_dof_container.dofs.keys():
                # Check if the current DOF is in the list of required DOFs
                if dof_name in dofs:
                    specific_dof_indices.append(index_counter)
                index_counter += 1
    
        return specific_dof_indices

# %% stiffness

    def Stiffness(self, omega):
        """
        Function to determine the global stiffness matrix of the element, 
        assembling it from its local element contributions.
        """
        # Get global and local DOF indices
        global_dof_indices = self.get_full_element_dof_indices_global()
        local_dof_indices = self.get_full_element_dof_indices_local()
    
        # Initialize full 12x12 stiffness matrix
        K_full = np.zeros((2 * Element.maxNdof, 2 * Element.maxNdof), dtype=complex)
    
        # Loop over all element types and accumulate stiffness contributions
        for element_type_name, element_type in self.element_types.items():
            # Get element-specific DOFs
            element_type_dofs = self.get_element_type_dofs(element_type)
            
            # Add element stiffness to full 12x12 matrix
            K_full[np.ix_(element_type_dofs, element_type_dofs)] += element_type.LocalStiffness(omega)
    
        # Apply the full rotation matrix to transform into the global coordinate system
        K_global_full = (self.R.T @ K_full) @ self.R  
    
        # Extract the relevant global DOFs **after** applying rotation
        K_global = K_global_full[np.ix_(global_dof_indices, global_dof_indices)]
    
        return K_global


# %% loads

    def AddDistributedLoad(self,**loads) -> None:
        '''
        Adds distributed loads to the element.
    
        Parameters
        ----------
        **loads : dict
            Keyword arguments where the key is the DOF (degree of freedom) and the value is the load magnitude.
        '''
        
        # Assign unpacked loads to the element's list of loads
        for dof, load in loads.items():
            if dof not in self.element_loads.keys():
                self.element_loads[dof] = load
            else:
                print(f'Load already added for DOF {dof}')

    def EvaluateDistributedLoad(self, element_loads, omega):
        '''
        Evaluates the distributed load
        '''
        # get the local dof indices - i.e. the indices of dofs present in the element per node
        local_dof_indices = self.get_full_element_dof_indices_global()
        
        # number of dofs in the current Element
        Ndof = len(local_dof_indices)
        
        # initialise local nodal load vector
        q_loc = np.zeros(Ndof, complex)
        
        
        for element_type_name, element_type in self.element_types.items():
            # get dofs of specific element type
            dofs = element_type.dofs
            
            # get the load into a vector belonging to only the dofs of the current element_type. Evaluate with omega if load is a lambda function
            q_evaluated = [
                element_loads.get(dof,0)(omega) if callable(element_loads.get(dof,0)) else element_loads.get(dof,0)
                for dof in dofs
                ]
            
            # pass through the loads that are specific for that element
            q_loc += self.FullDistributedLoad(element_type, q_evaluated, omega)[np.ix_(local_dof_indices)]
        
        # TODO - check whether 
        q_glob = self.R[np.ix_(local_dof_indices,local_dof_indices)] @ q_loc
        
        return q_glob   

    def FullDistributedLoad(self, element_type, q, omega):
        '''
        Get the nodal loads from the distributed load per element type and put in the full load vector
        '''
        # get the element_type dofs
        element_type_dofs = self.get_element_type_dofs(element_type)
        
        # initialise maxNdof empty complex vector 
        q_full = np.zeros(2*Element.maxNdof,complex)
        
        # set directly
        q_full[np.ix_(element_type_dofs)] = element_type.LocalDistributedLoad(q, omega)

        
        return q_full             

#%% properties of the Element    
            
    def geometrical_properties(self,alpha = 0):
        '''
        Determines the following geometrical properties of the element:
            - length
            - full rotation matrix (12x12)
    
        the full displacement vector would be as follows:            
            u = [u_x_l, u_z_l, u_y_l, phi_x_l, phi_z_l, phi_y_l, u_x_r, u_z_r, u_y_r, phi_x_r, phi_z_r, phi_y_r]            
            
        Parameters
        ----------
        alpha : rotation in degrees of the cross-section. I ASSUME A ROTATION OF 0 MEANS THE CROSS SECTION IS PERFECTLY VERTICAL? NEED TO CHECK
        '''
        
        ## Calculate length of beam (L) and orientation 
        xl, zl, yl = self.nodes[0].get_coords() # x, z and y of left node

        xr, zr, yr = self.nodes[1].get_coords() # x, z and y of left node

        # get the length of the element
        l = np.sqrt((xr-xl)**2 + (zr-zl)**2 + (yr-yl)**2)           
        self.L = l
        
        # get the angles 
        Cx = (xr-xl)/l
        Cz = (zr-zl)/l
        Cy = (yr-yl)/l
        Cxy = np.sqrt(Cx**2 + Cy**2)
        
        ## Determining rotation if alpha is not given
        
        if alpha is None:
            # get the rotation about the local x-axis according to nonstructural node
            # no nonstructnode is given, thus create one on the same x,y plane and above and outside of it - NEED DEFINITION OF NONSTRUCTNODE
            nonstructnode = []
                
            # first vector for determining global plane
            P1 = np.array(nonstructnode) - np.array([xl, zl, yl])  
            # second vector for determining global plane
            P2 = np.array([xr, zr, yr]) - np.array([xl, zl, yl])  
            # normal of the plane determined by nonstructural node
            normal1 = np.cross(P1,P2)          
            # vectors for plane in local unrotated plane
            P3 = np.array([1, 0, 0])                     
            P4 = np.array([1, 1, 0])                     
            
            # calculate the rotation
            if Cxy == 0:  # if element is placed vertically
                YY = np.array([[0, Cz, 0],
                               [-Cz, 0, 0],
                               [0, 0, 1]])
            else:
                # first 2 rotation matrices for basic element rotation neglecting rotation
                # about the local x-axis
                Rb = np.array([[Cx/Cxy, 0, Cz/Cxy],
                               [0, 1, 0],
                               [-Cy/Cxy, 0, Cx/Cxy]])
                Rz = np.array([[Cxy, Cz, 0],
                               [-Cz, Cxy, 0],
                               [0, 0, 1]])      
                YY = Rz @ Rb
                
            # rotate the basic vectors
            P3 = YY.T @ P3; P3 = P3.T  
            P4 = YY.T @ P4; P4 = P4.T
            # determine normal of the element plane without rotation
            normal2 = np.cross(P4,P3)  
        
            # now calculate the angle between the 2 planes:
            alpha = np.degrees(np.arctan2(np.linalg.norm(np.cross(normal1,normal2)), np.dot(normal1,normal2)))
            
            if alpha != 0:
                print(f'alpha = {alpha} !')
        
        # and do the rotation about all axes
        if Cxy == 0: # check whether member will be placed vertically
            Y = np.array([[0, Cz, 0], [-Cz*np.cos(np.radians(alpha)), 0, np.sin(np.radians(alpha))], [Cz*np.sin(np.radians(alpha)), 0, np.cos(np.radians(alpha))]])
        else:
            Y = np.array([[Cx, Cz, Cy], [(-Cx*Cz*np.cos(np.radians(alpha))-Cy*np.sin(np.radians(alpha)))/Cxy, Cxy*np.cos(np.radians(alpha)), (-Cz*Cy*np.cos(np.radians(alpha))+Cx*np.sin(np.radians(alpha)))/Cxy], [(Cx*Cz*np.sin(np.radians(alpha))-Cy*np.cos(np.radians(alpha)))/Cxy, -Cxy*np.sin(np.radians(alpha)), (Cz*Cy*np.sin(np.radians(alpha))+Cx*np.cos(np.radians(alpha)))/Cxy]])
    
        # full rotation matrix (12x12)
        self.R = np.block([[Y, np.zeros((3,9))], [np.zeros((3,3)), Y, np.zeros((3,6))], [np.zeros((3,6)), Y, np.zeros((3,3))], [np.zeros((3,9)), Y]])
        
# %% Setting / removing a section        
    def SetSection(self, element_type, props):
        '''
        This function serves to set the elements, we will have to think how we do this properly but this is an initial set-up
        
        '''
        # TODO - need to handle errors correctly when not the correct parameters are given
        
        # add L to props
        props['L'] = self.L
        
        try:
            #  try to load the element from the element factory                
            element = ElementFactory.CreateElement(element_type, **props)
            
            # Add local DOFs to each node's local DOFContainer
            for node in self.nodes:
                local_container = self.local_dof_container[node.id]
                for dof_name in element.dofs:
                    if not local_container.has_dof(dof_name):
                        local_container.set_dof(dof_name)
            
            # check influence of the local dofs of the element on the global dofs of the current node
            self.apply_global_constraints_to_local_dofs()
            
            # store the element
            self.element_types[element_type] = element    
        except Exception as e:
            print(f'Exception occurred - {e}')
        else:
            print(f'Successfully added element of type: {element_type} to Element {self.id}')
    
    def RemoveSection(self,element_type):
        '''
        Does what it says it does..
        
        Returns
        -------
        
        '''        
        try:
            del self.element_types[element_type]
        except Exception as e:
            print(f'An error has occurred whilst trying to remove a section - {e}')

                
# %% Prescribing dofs   
    def fix_dof(self, node, *dofs):
        """
        Fixes specified DOFs at a given node in the element.
    
        Parameters
        ----------
        node : Node
            The node where DOFs are to be fixed.
        *dofs : str
            DOFs to be fixed at the node.
        """
        
        self.prescribe_dof(node, **{dof_name: 0 for dof_name in dofs})
                
    def free_dof(self, node, *dofs):
        """
        Frees specified DOFs at a given node in the element.
    
        Parameters
        ----------
        node : Node
            The node where DOFs are to be freed.
        *dofs : str
            DOFs to be freed at the node.
        """
        self.prescribe_dof(node, **{dof_name: None for dof_name in dofs})
                
        
    def prescribe_dof(self, node, **dofs):
        """
        Prescribes specified DOFs at a given node in the element to given values.
    
        Parameters
        ----------
        node : Node
            The node where DOFs are to be prescribed.
        **dofs : dict
            DOFs and their values to be prescribed at the node.
    
        Raises
        ------
        ValueError
            If any of the specified DOFs cannot be added due to lack of influence by local DOFs.
        """
        try:
            dof_container = self.dof_containers[node.id]  # Access the node's DOFContainer
            for dof_name, value in dofs.items():
                # Check if the DOF is already in the node's DOFContainer
                if dof_container.has_dof(dof_name):
                    # Update the DOF with the new value
                    self._update_dof(node, dof_name, value)
                else:
                    # Get list of global dofs affected by local dofs
                    global_dofs = self.map_local_to_global_dofs(self.local_dof_container[node.id].dofs.keys())
                    # Flatten the list of global DOFs influenced by local DOFs
                    influenced_global_dofs = [gdof for gdofs in global_dofs.values() for gdof in gdofs]
                    
                    # Check if the global DOF is influenced by local DOFs
                    if dof_name in influenced_global_dofs:
                        self._handle_missing_dof(node, dof_name, value)
                    else:
                        print(f"DOF '{dof_name}' cannot be added, as it is not affected by local DOFs in the current element configuration.")
        except ValueError as e:
            print(e)
    
    
    def _update_dof(self, node, dof_name, value):
        """
        Updates the DOF for a given node, also updating local DOFs if necessary.
    
        Parameters
        ----------
        node : Node
            The node to update.
        dof_name : str
            The DOF to be updated.
        value : Any
            The value to prescribe to the DOF.
        """
        # Update the global DOF in the element's DOFContainer
        dof_container = self.dof_containers[node.id]
        dof_container.set_dof(dof_name, value=value)
    
        # Update the local DOFs based on the new global DOF
        local_dofs = self.map_global_to_local_dofs([dof_name])
    
        for local_dof_name in local_dofs.get(dof_name, []):
            if self.local_dof_container[node.id].has_dof(local_dof_name):
                self.local_dof_container[node.id].set_dof_value(local_dof_name, value=value)

    def couple_dof(self, node, *dofs):
        """
        Ensures that the specified DOFs are monolithically connected to the node.
        This means the DOF will be fully constrained across all elements at this node.
    
        Parameters
        ----------
        node : Node
            The node where the DOFs should be coupled.
        *dofs : str
            DOFs that should be monolithically connected.
        """
        for dof_name in dofs:
            if dof_name in self.dof_containers[node.id].dofs:
                self.constraint_types[node.id][dof_name] = "monolithic"
    
    
    def decouple_dof(self, node, *dofs):
        """
        Allows the specified DOFs to move freely in the element, preventing monolithic connection.
    
        Parameters
        ----------
        node : Node
            The node where the DOFs should be decoupled.
        *dofs : str
            DOFs that should be independent in the element.
        """
        for dof_name in dofs:
            if dof_name in self.dof_containers[node.id].dofs:
                self.constraint_types[node.id][dof_name] = "independent"

    
    def _handle_missing_dof(self, node, dof_name, value):
        """
        Handles the case where a global DOF is not present in the element but can be influenced by local DOFs.
    
        Parameters
        ----------
        node : Node
            The node to update.
        dof_name : str
            The missing DOF.
        value : Any
            The value to prescribe to the DOF.
        """
        # Check for user input
        response = input(f"DOF '{dof_name}' is not in the element. Would you like to add it? (yes/no): ").strip().lower()
        if response == 'no':
            return
    
        # If answer is yes, add missing DOF if influenced by local DOFs
        dof_container = self.dof_containers[node.id]
        dof_container.set_dof(dof_name, value=value)
        print(f"DOF '{dof_name}' added to the element and prescribed to {value}.")
    
        # Update local DOFs based on the new global DOF
        local_dofs = self.map_global_to_local_dofs([dof_name])
    
        for local_dof_name in local_dofs.get(dof_name, []):
            if self.local_dof_container[node.id].has_dof(local_dof_name):
                self.local_dof_container[node.id].set_dof_value(local_dof_name, value=value)
        print(f"Local DOFs updated to reflect new global DOF '{dof_name}'.")
         

    def update_node_dofs(self, node, changes):
        """
        Updates the DOFs of a given node in the element based on changes.
    
        Parameters
        ----------
        node : Node
            The node whose DOFs are to be updated.
        changes : dict
            Dictionary of DOFs and their new values.
        """
        if node in self.nodes:
            dof_container = self.dof_containers[node.id]
            for dof_name, value in changes.items():
                if dof_container.has_dof(dof_name):
                    dof_container.set_dof(dof_name, value=value)
                    
# %% Element Displacements

    def Displacements(self, u_nodes_global, omega, num_points=20):
        """
        Gets the displacements over the local axis of the element evaluated at num_points.
    
        Result is structured as (example in 2D config):
            u_elem = [u_x(s)
                      u_z(s)
                      phi_y(s)]
    
        Where:
            - u_x(s) = axial displacement 
            - u_z(s) = transverse displacement
            - phi_y(s) = rotation about y axis
    
        Input:
            u_nodes_global: [u_x_left
                             u_z_left
                             phi_y_left
                             u_x_right
                             u_z_right
                             phi_y_right]
        """
        
        # Get all element dof indices present based on dof mapping
        local_dof_indices = self.get_full_element_dof_indices_local()
        global_dof_indices = self.get_full_element_dof_indices_global()
        Ndof = len(local_dof_indices)
    
        # Convert the global nodal displacements to the local element coordinate system
        u_local = self.R[np.ix_(local_dof_indices, global_dof_indices)] @ u_nodes_global
    
        # Initialize empty list to hold displacements for each DOF at specified points
        u_elem = np.array([np.zeros(num_points, dtype=complex) for _ in range(Element.maxNdof)])
    
        # Loop over all element types
        for element_type_name, element_type in self.element_types.items():
            # Get the DOFs of the element type in the context of the nodes
            dofs = element_type.dofs
            specific_dof_indices = self.get_specific_dof_indices_local(dofs)
    
            # Calculate local displacements for the specific element type
            u_elem_contribution = element_type.LocalElementDisplacements(
                u_local[np.ix_(specific_dof_indices)], omega, num_points
            )
    
            # Add the contribution of each element type to the full local element displacements
            for i, dof in enumerate(dofs):
                global_dof_index = self.dof_mapping[dof]
                if u_elem_contribution[i] is not None:
                    u_elem[global_dof_index] += u_elem_contribution[i]
    
        u_global = self.R[:6,:6].T @ u_elem
    
        return u_global
    
# %% Element Forces
    def Forces(self, u_nodes_global, omega, num_points=20):
        """
        Gets the Forces over the local axis of the element evaluated at num_points.
    
        Result is structured as (example in 2D config):
            u_elem = [Nx(s)
                      Vz(s)
                      Myy(s)]
    
        Where:
            - Nx(s) = axial force along x axis 
            - Vz(s) = shear force along z axis 
            - Myy(s) = bending moment about y-y axis
    
        Input:
            u_nodes_global: [u_x_left
                             u_z_left
                             phi_y_left
                             u_x_right
                             u_z_right
                             phi_y_right]
        """
        # Get all element dof indices present based on dof mapping
        local_dof_indices = self.get_full_element_dof_indices_local()
        global_dof_indices = self.get_full_element_dof_indices_global()
        
        # Convert the global nodal displacements to the local element coordinate system
        u_local = self.R[np.ix_(local_dof_indices, global_dof_indices)] @ u_nodes_global
        
        # Initialize empty list to hold displacements for each DOF at specified points
        F_elem = np.array([np.zeros(num_points, dtype=complex) for _ in range(Element.maxNdof)])
        
        # Loop over all element types
        for element_type_name, element_type in self.element_types.items():
            # Get the DOFs of the element type in the context of the nodes
            dofs = element_type.dofs
            specific_dof_indices = self.get_specific_dof_indices_local(dofs)
            
            # Calculate local displacements for the specific element type
            F_elem_contribution = element_type.LocalElementForces(
                u_local[np.ix_(specific_dof_indices)], omega, num_points
            )
            
            # Add the contribution of each element type to the full local elment displacements
            for i, dof in enumerate(dofs):
                global_dof_index = self.dof_mapping[dof]
                if F_elem_contribution[i] is not None: 
                    F_elem[global_dof_index] += F_elem_contribution[i]
                    
        F_global =  F_elem
   
        return F_global
    
# %% Element Stresses
    def Stresses(self, u_nodes_global, omega, num_points=20):
            """
            Gets the Stresses over the local axis of the element evaluated at num_points.
        
            Result is structured as (example in 2D config):
                sigma_elem = [sigmaxx(s)
                              tau(s)]
        
            Where:
                - sigmaxx(s) = axial stress along x axis 
                - tau(s) = averaged shear stress
        
            Input:
                u_nodes_global: [u_x_left
                                 u_z_left
                                 phi_y_left
                                 u_x_right
                                 u_z_right
                                 phi_y_right]
            """
            # Get all element dof indices present based on dof mapping
            local_dof_indices = self.get_full_element_dof_indices_local()
            global_dof_indices = self.get_full_element_dof_indices_global()
            
            # Convert the global nodal displacements to the local element coordinate system
            u_local = self.R[np.ix_(local_dof_indices, global_dof_indices)] @ u_nodes_global
            
            # Initialize empty list to hold displacements for each DOF at specified points
            sigma_elem = np.array([np.zeros(num_points, dtype=complex) for _ in range(Element.maxNdof)])
            
            # Loop over all element types
            for element_type_name, element_type in self.element_types.items():
                # Get the DOFs of the element type in the context of the nodes
                dofs = element_type.dofs
                specific_dof_indices = self.get_specific_dof_indices_local(dofs)
                
                # Calculate local displacements for the specific element type
                sigma_elem_contribution = element_type.LocalElementStresses(
                    u_local[np.ix_(specific_dof_indices)], omega, num_points
                )
                
                # Add the contribution of each element type to the full local elment displacements
                for i, dof in enumerate(dofs):
                    global_dof_index = self.dof_mapping[dof]
                    if sigma_elem_contribution[i] is not None: 
                        sigma_elem[global_dof_index] += sigma_elem_contribution[i]
                        
            sigma_global = sigma_elem
            
            return sigma_global
        
# %% help functions    
    def create_local_dofs_vector(self, node):
        """
        Creates a vector with ones where the local DOFs are present and zeros elsewhere.

        Returns
        -------
        numpy.ndarray
            A vector of size equal to the number of DOFs, with ones indicating the presence of local DOFs.
        """
        dof_vector = np.zeros(len(self.dof_mapping))
        for dof_name in self.local_dof_container[node.id].dofs:
            if dof_name in self.dof_mapping:
                index = self.dof_mapping[dof_name]
                dof_vector[index] = 1
        return dof_vector

    def local_to_global_dof(self, local_dof_vector):
        """
        Transforms a local DOF vector into the global DOF vector using the rotation matrix.
    
        Parameters
        ----------
        local_dof_vector : numpy.ndarray
            The local DOF vector.
    
        Returns
        -------
        global_dof_vector : numpy.ndarray
            The global DOF vector.            
        """
        
        global_dof_vector = self.R[:6,:6].T @ local_dof_vector 
        return global_dof_vector
    
    def global_to_local_dof(self, global_dof_vector):
        """
        Transforms a global DOF vector into the local DOF vector using the inverse of the rotation matrix.
    
        Parameters
        ----------
        global_dof_vector : numpy.ndarray
            The global DOF vector.
    
        Returns
        -------
        local_dof_vector : numpy.ndarray
            The local DOF vector.
        """
        
        # Apply the rotation matrix transpose (which is the inverse for orthogonal matrices)
        local_dof_vector = self.R[:6, :6] @ global_dof_vector
        return local_dof_vector
    
    def map_local_to_global_dofs(self,local_dof_names):
        """
        Maps local DOFs to affected global DOFs.
    
        Parameters
        ----------
        local_dof_names : list or iterable of str
            Names of local DOFs.
    
        Returns
        -------
        dict
            Mapping from local DOF names to lists of affected global DOF names.
        """
        affected_global_dofs = {}
        for local_dof_name in local_dof_names:
            local_dof_vector = np.zeros(len(self.dof_mapping))
            local_dof_vector[self.dof_mapping[local_dof_name]] = 1
    
            # Transform to global DOF space
            global_dof_vector = self.local_to_global_dof(local_dof_vector)
    
            # Identify the affected global DOFs
            affected_indices = np.nonzero(global_dof_vector)[0]
            global_dof_names = [self.reverse_dof_mapping[i] for i in affected_indices]
            affected_global_dofs[local_dof_name] = global_dof_names
        return affected_global_dofs
    
    def map_global_to_local_dofs(self, global_dof_names):
        """
        Maps global DOFs to affected local DOFs.
    
        Parameters
        ----------
        global_dof_names : list
            List of global DOF names.
    
        Returns
        -------
        dict
            Mapping from global DOF names to lists of affected local DOF names.
        """
        affected_local_dofs = {}
        for global_dof_name in global_dof_names:
            global_dof_vector = np.zeros(len(self.dof_mapping))
            global_dof_vector[self.dof_mapping[global_dof_name]] = 1
    
            # Transform to local DOF space
            local_dof_vector = self.global_to_local_dof(global_dof_vector)
    
            # Identify the affected local DOFs
            affected_indices = np.nonzero(local_dof_vector)[0]
            local_dof_names = [self.reverse_dof_mapping[i] for i in affected_indices]
            affected_local_dofs[global_dof_name] = local_dof_names
        return affected_local_dofs
    
    def apply_global_constraints_to_local_dofs(self):
        """
        Maps global DOF constraints to local DOFs and applies these constraints to the local DOFs
        of the element.
        """
        for node in self.nodes:
            # Access global DOFs from the node's DOFContainer
            global_container = node.dof_container
            global_dof_names = [dof_name for dof_name in global_container.dofs]
            local_dof_map = self.map_global_to_local_dofs(global_dof_names)
            
            local_dof_container = self.local_dof_container[node.id]
            
            # Apply constraints from global DOFs to local DOFs
            for global_dof_name, local_dof_names in local_dof_map.items():
                global_dof = global_container.get_dof(global_dof_name)
                global_dof_value = global_dof.value
    
                if global_dof_value is not None:
                    for local_dof_name in local_dof_names:
                        if local_dof_container.has_dof(local_dof_name):
                            local_dof_container.set_dof(local_dof_name, value=global_dof_value)

    def apply_global_dof_change(self, node, global_dof_name, global_dof_value):
        """
        Applies the change in a specific global DOF to both the global and local DOFs
        of the element.
        
        Parameters
        ----------
        node : Node
            The node whose global DOF is being changed.
        global_dof_name : str
            The name of the global DOF that has changed.
        global_dof_value : float
            The new value of the global DOF (fixed, free, or prescribed).
        """
        # Update the global DOF in the node's DOFContainer
        dof_container = self.dof_containers[node.id]
        if dof_container.has_dof(global_dof_name):
            dof_container.set_dof_value(global_dof_name, global_dof_value)
            print(f"Global DOF '{global_dof_name}' for node {node.id} updated to {global_dof_value}.")
        
        # Map the changed global DOF to the local DOFs
        local_dof_map = self.map_global_to_local_dofs([global_dof_name])
        
        local_dof_container = self.local_dof_container[node.id]
        
        # Apply the change to the local DOFs
        for local_dof_name in local_dof_map.get(global_dof_name, []):
            if local_dof_container.has_dof(local_dof_name):
                local_dof_container.set_dof_value(local_dof_name, global_dof_value)
                print(f"Local DOF '{local_dof_name}' updated to {global_dof_value}.")
    
    def check_and_handle_global_dofs(self,dofs):
        """
        Checks whether the affected global DOFs are present in self.dofs. If not, prompts for user action.

        Returns
        -------
        bool
            True if all affected global DOFs are present or user chooses to continue, False otherwise.
        """
        affected_dofs = self.check_global_dof_influence(dofs)
        missing_dofs = []

        for local_dof, global_dofs in affected_dofs.items():
            for global_dof in global_dofs:
                for node_id, node_dofs in self.dofs.items():
                    if global_dof not in node_dofs:
                        missing_dofs.append((node_id, global_dof))

        # if all dofs present, do nothing and return 
        if not missing_dofs:
            return 

        print(f"Missing global DOFs: {missing_dofs}")
        for node_id, missing_dof in missing_dofs:    
           # Check if the missing DOF is present in the other node's DOFs
           other_node_id = next(key for key in self.dofs.keys() if key != node_id)
           if missing_dof in self.dofs[other_node_id]:
               print(f"The missing DOF {missing_dof} at node_id {node_id} is present in the other node_id {other_node_id}.")
               action = input(f"The global DOF {missing_dof} at node_id {node_id} is missing. It is present in node_id {other_node_id}. Do you want to do something?: ").strip().lower()
           else:
               action = input(f"The global DOF {missing_dof} at node_id {node_id} is missing. Do you want to do something?: ").strip().lower()
           
           if action == 'yes':
               return True
    
        return False


# -*- coding: utf-8 -*-
"""
Created on Mon Mar 18 19:52:53 2024

@author: rensv
"""

# import necessary libraries
import matplotlib.pyplot as plt
import numpy as np

# %% Class definition
class StructurePlotter:
    def __init__(self):
        pass
    
    def PlotNodes(self, nodes):
        plt.figure(figsize=(10, 6))
        for node in nodes:
            # marker = 'o' if len(node.constrained_dofs) == 0 else '^'
            # color = 'blue' if len(node.constrained_dofs) == 0 else 'red'
            # plt.scatter(node.x, node.z, color=color, marker=marker, label=f'{"Free" if len(node.constrained_dofs) == 0 else "Fixed"} Node' if f'{"Free" if len(node.constrained_dofs) == 0 else "Fixed"} Node' not in plt.gca().get_legend_handles_labels()[1] else "")
            plt.scatter(node.x, node.z, color='red', marker='o', label=f'Node: {node.id}')
            plt.text(node.x+0.02, node.z+0.02, f'{node.id}', fontsize=9, ha='right')

    def PlotElements(self, elements):
        for element in elements:
            x_values = [element.nodes[0].x, element.nodes[1].x]
            z_values = [element.nodes[0].z, element.nodes[1].z]
            plt.plot(x_values, z_values, 'k-', linewidth=2, label=f'Element: {element.id}' if 'Element' not in plt.gca().get_legend_handles_labels()[1] else "")
            mid_x = sum(x_values) / 2
            mid_z = sum(z_values) / 2
            plt.text(mid_x, mid_z+0.02, element.id, fontsize=9, color='black')
            
    def PlotDisplacements(self, elements, displacements, scale=1.0):
        """
        Plots the displaced shape of the structure based on the provided displacements.

        Parameters
        ----------
        elements : list of Element
            The list of elements in the structural system.
        displacements : list of numpy.ndarray
            List of displacements for each element as returned by `Analysis.ElementDisplacements`.
        scale : float
            Scale factor for displacements to visualize the deformations clearly.
        num_points : int
            Number of points along each element length for plotting the displaced shape.
        """
        plt.figure(figsize=(10, 8))
        plt.title('Real and Imaginary Parts of Displaced Structure')
        plt.xlabel('X Coordinate')
        plt.ylabel('Z Coordinate')
        plt.grid(True)
        plt.axis('equal')

        # Plot the original structure (nodes and elements)
        for node in [node for element in elements for node in element.nodes]:
            plt.scatter(node.x, node.z, color='red', marker='o', label='Node' if 'Node' not in plt.gca().get_legend_handles_labels()[1] else "")
            plt.text(node.x+0.02, node.z+0.02, f'{node.id}', fontsize=9, ha='right')
        for element in elements:
            x_values = [element.nodes[0].x, element.nodes[1].x]
            z_values = [element.nodes[0].z, element.nodes[1].z]
            plt.plot(x_values, z_values, 'k--', linewidth=2, label='Element' if 'Element' not in plt.gca().get_legend_handles_labels()[1] else "")

        displaced_nodes_real = {}
        displaced_nodes_imag = {}

        # Plot the displaced elements (real and imaginary parts)
        for element in elements:
            # Extract the original coordinates of the nodes
            x0 = element.nodes[0].x
            z0 = element.nodes[0].z
            x1 = element.nodes[1].x
            z1 = element.nodes[1].z

            # Get the displacement components (assuming u_elem has shape (6, num_points))
            u_elem = displacements[element.id]
            u_x_real = np.real(u_elem[0, :])
            u_z_real = np.real(u_elem[1, :])
            u_x_imag = np.imag(u_elem[0, :])
            u_z_imag = np.imag(u_elem[1, :])
            
            num_points = len(u_x_real)
            # Create an array of points along the length of the element
            s = np.linspace(0, 1, num_points)

            # Calculate the original positions along the element's length
            x = x0 + (x1 - x0) * s
            z = z0 + (z1 - z0) * s

            # Add the displacements scaled by the specified factor
            x_disp_real = x + scale * u_x_real
            z_disp_real = z + scale * u_z_real
            x_disp_imag = x + scale * u_x_imag
            z_disp_imag = z + scale * u_z_imag

            # Plot the deformed shape of the element (real and imaginary parts)
            plt.plot(x_disp_real, z_disp_real, 'b-', linewidth=1.5, label='Real Part of Displaced Element' if 'Real Part of Displaced Element' not in plt.gca().get_legend_handles_labels()[1] else "")
            plt.plot(x_disp_imag, z_disp_imag, 'g--', linewidth=1.5, label='Imaginary Part of Displaced Element' if 'Imaginary Part of Displaced Element' not in plt.gca().get_legend_handles_labels()[1] else "")

            # Save the displaced coordinates of the nodes (start and end points) for both real and imaginary parts
            if element.nodes[0].id not in displaced_nodes_real:
                displaced_nodes_real[element.nodes[0].id] = (x_disp_real[0], z_disp_real[0])
            if element.nodes[1].id not in displaced_nodes_real:
                displaced_nodes_real[element.nodes[1].id] = (x_disp_real[-1], z_disp_real[-1])

            if element.nodes[0].id not in displaced_nodes_imag:
                displaced_nodes_imag[element.nodes[0].id] = (x_disp_imag[0], z_disp_imag[0])
            if element.nodes[1].id not in displaced_nodes_imag:
                displaced_nodes_imag[element.nodes[1].id] = (x_disp_imag[-1], z_disp_imag[-1])

        # Plot the displaced nodes (real and imaginary parts)
        for node_id, (x_disp_real, z_disp_real) in displaced_nodes_real.items():
            plt.scatter(x_disp_real, z_disp_real, color='blue', marker='x', label='Real Part of Displaced Node' if 'Real Part of Displaced Node' not in plt.gca().get_legend_handles_labels()[1] else "")
            plt.text(x_disp_real+0.02, z_disp_real+0.02, f'{node_id}', fontsize=9, ha='right', color='blue')

        for node_id, (x_disp_imag, z_disp_imag) in displaced_nodes_imag.items():
            plt.scatter(x_disp_imag, z_disp_imag, color='green', marker='x', label='Imaginary Part of Displaced Node' if 'Imaginary Part of Displaced Node' not in plt.gca().get_legend_handles_labels()[1] else "")
            plt.text(x_disp_imag+0.02, z_disp_imag+0.02, f'{node_id}', fontsize=9, ha='right', color='green')

        # Add legend and show the plot
        plt.legend()
        plt.show()

    def Plotmoments(self, elements, forces, scale=1e-8): # TODO: check the direction of moment, shear force lines
        
        plt.figure(figsize=(10, 8))
        plt.title('Bending moments plot')
        plt.axis('off')

        # Plot the original structure (nodes and elements)
        for node in [node for element in elements for node in element.nodes]:
            plt.scatter(node.x, node.z, color='red', marker='o', label='Node' if 'Node' not in plt.gca().get_legend_handles_labels()[1] else "")
            plt.text(node.x+0.02, node.z+0.02, f'{node.id}', fontsize=9, ha='right')
        for element in elements:
            x_values = [element.nodes[0].x, element.nodes[1].x]
            z_values = [element.nodes[0].z, element.nodes[1].z]
            plt.plot(x_values, z_values, 'k--', linewidth=2, label='Element' if 'Element' not in plt.gca().get_legend_handles_labels()[1] else "")

        mmt_nodes_real = {}
        mmt_nodes_imag = {}

        # Plot the displaced elements (real and imaginary parts)
        for element in elements:
            # Extract the original coordinates of the nodes
            x0 = element.nodes[0].x
            z0 = element.nodes[0].z
            x1 = element.nodes[1].x
            z1 = element.nodes[1].z
            
            # Get the displacement components (assuming u_elem has shape (6, num_points))
            f_elem = forces[element.id]
            m_real = np.real(f_elem[-1, :])
            m_imag = np.imag(f_elem[-1, :])
            
            # m_real = m_real/np.max(np.abs(m_real)) # need to be removed after the development of scaling factor
            # m_imag = m_imag
            
            num_points = len(m_real)
            # Create an array of points along the length of the element
            s = np.linspace(0, 1, num_points)

            # Calculate the original positions along the element's length
            x = x0 + (x1 - x0) * s
            z = z0 + (z1 - z0) * s
          # Compute perpendicular direction for offsetting the moment line
            dx = x1 - x0
            dz = z1 - z0
            length = np.sqrt(dx**2 + dz**2)
            perp_x = -dz / length  # Unit normal in X
            perp_z = dx / length   # Unit normal in Z          
            
            moment_x = x + scale * m_real * perp_x
            moment_z = z + scale * m_real * perp_z
            
            moment_x_imag = x + scale * m_imag * perp_x
            moment_z_imag = z + scale * m_imag * perp_z
            
            # Plot the moment line
            plt.plot(moment_x, moment_z, 'b-', linewidth=1, label='Real Part of Bending Moment' if element.id == elements[0].id else "")
            plt.plot(moment_x_imag, moment_z_imag, 'g--', linewidth=1, label='Imaginary Part of Bending Moment' if element.id == elements[0].id else "")
            
            # Connect moment diagram back to the structure
            plt.plot([x[0], moment_x[0]], [z[0], moment_z[0]], 'b--', linewidth=1)
            plt.plot([x[-1], moment_x[-1]], [z[-1], moment_z[-1]], 'b--', linewidth=1)
            plt.plot([x[0], moment_x_imag[0]], [z[0], moment_z_imag[0]], 'g--', linewidth=1)
            plt.plot([x[-1], moment_x_imag[-1]], [z[-1], moment_z_imag[-1]], 'g--', linewidth=1)
        # Add legend and show the plot
        plt.legend()
        plt.show()    

    def Plotaxialforces(self, elements, forces, scale=1e-8):
        
        plt.figure(figsize=(10, 8))
        plt.title('Axial Force Plot')
        plt.axis('off')
  
        # Plot the original structure (nodes and elements)
        for node in [node for element in elements for node in element.nodes]:
            plt.scatter(node.x, node.z, color='red', marker='o', label='Node' if 'Node' not in plt.gca().get_legend_handles_labels()[1] else "")
            plt.text(node.x+0.02, node.z+0.02, f'{node.id}', fontsize=9, ha='right')
        for element in elements:
            x_values = [element.nodes[0].x, element.nodes[1].x]
            z_values = [element.nodes[0].z, element.nodes[1].z]
            plt.plot(x_values, z_values, 'k--', linewidth=2, label='Element' if 'Element' not in plt.gca().get_legend_handles_labels()[1] else "")
  
        mmt_nodes_real = {}
        mmt_nodes_imag = {}
  
        # Plot the displaced elements (real and imaginary parts)
        for element in elements:
            # Extract the original coordinates of the nodes
            x0 = element.nodes[0].x
            z0 = element.nodes[0].z
            x1 = element.nodes[1].x
            z1 = element.nodes[1].z
            
            # Get the displacement components (assuming u_elem has shape (6, num_points))
            f_elem = forces[element.id]
            N_real = np.real(f_elem[0, :])
            N_imag = np.imag(f_elem[0, :])
            
            # N_real = N_real/np.max(np.abs(N_real))
            
            num_points = len(N_real)
            # Create an array of points along the length of the element
            s = np.linspace(0, 1, num_points)
  
            # Calculate the original positions along the element's length
            x = x0 + (x1 - x0) * s
            z = z0 + (z1 - z0) * s
          # Compute perpendicular direction for offsetting the moment line
            dx = x1 - x0
            dz = z1 - z0
            length = np.sqrt(dx**2 + dz**2)
            perp_x = dz / length  # Unit normal in X
            perp_z = dx / length   # Unit normal in Z          
            
            axialforce_x = x + scale * N_real * perp_x
            axialforce_z = z - scale * N_real * perp_z
            
            axialforce_x_imag = x + scale * N_imag * perp_x
            axialforce_z_imag = z - scale * N_imag * perp_x
            
            # Plot the moment line
            plt.plot(axialforce_x, axialforce_z, 'b-', linewidth=1, label='Real Part of Axial Force' if element.id == elements[0].id else "")
            plt.plot(axialforce_x_imag, axialforce_z_imag, 'g--', linewidth=1, label='Imaginary Part of Axial Force' if element.id == elements[0].id else "")
        
            # Connect moment diagram back to the structure
            plt.plot([x[0], axialforce_x[0]], [z[0], axialforce_z[0]], 'b--', linewidth=1)
            plt.plot([x[-1], axialforce_x[-1]], [z[-1], axialforce_z[-1]], 'b--', linewidth=1)
            plt.plot([x[0], axialforce_x_imag[0]], [z[0], axialforce_z_imag[0]], 'g--', linewidth=1)
            plt.plot([x[-1], axialforce_x_imag[-1]], [z[-1], axialforce_z_imag[-1]], 'g--', linewidth=1)
  
        # Add legend and show the plot
        plt.legend()
        plt.show()
        
    def Plotshearforces(self, elements, forces, scale=1e-8):
        
        plt.figure(figsize=(10, 8))
        plt.title('Shear Force Plot')
        plt.axis('off')
  
        # Plot the original structure (nodes and elements)
        for node in [node for element in elements for node in element.nodes]:
            plt.scatter(node.x, node.z, color='red', marker='o', label='Node' if 'Node' not in plt.gca().get_legend_handles_labels()[1] else "")
            plt.text(node.x+0.02, node.z+0.02, f'{node.id}', fontsize=9, ha='right')
        for element in elements:
            x_values = [element.nodes[0].x, element.nodes[1].x]
            z_values = [element.nodes[0].z, element.nodes[1].z]
            plt.plot(x_values, z_values, 'k--', linewidth=2, label='Element' if 'Element' not in plt.gca().get_legend_handles_labels()[1] else "")
  
        sf_nodes_real = {}
        sf_nodes_imag = {}
  
        # Plot the displaced elements (real and imaginary parts)
        for element in elements:
            # Extract the original coordinates of the nodes
            x0 = element.nodes[0].x
            z0 = element.nodes[0].z
            x1 = element.nodes[1].x
            z1 = element.nodes[1].z
            
            # Get the displacement components (assuming u_elem has shape (6, num_points))
            f_elem = forces[element.id]
            V_real = np.real(f_elem[1, :])
            V_imag = np.imag(f_elem[1, :])
            
            num_points = len(V_real)
            # Create an array of points along the length of the element
            s = np.linspace(0, 1, num_points)
      
            # Calculate the original positions along the element's length
            x = x0 + (x1 - x0) * s
            z = z0 + (z1 - z0) * s
          # Compute perpendicular direction for offsetting the moment line
            dx = x1 - x0
            dz = z1 - z0
            length = np.sqrt(dx**2 + dz**2)
            perp_x = dz / length  # Unit normal in X
            perp_z = dx / length   # Unit normal in Z          
            
            shearforce_x = x - scale * V_real * perp_x
            shearforce_z = z + scale * V_real * perp_z
            
            shearforce_x_imag = x - scale * V_imag * perp_x
            shearforce_z_imag = z + scale * V_imag * perp_z
            
            # Plot the moment line
            plt.plot(shearforce_x, shearforce_z, 'b-', linewidth=1, label='Real Part of Shear Force' if element.id == elements[0].id else "")
            plt.plot(shearforce_x_imag, shearforce_z_imag, 'g--', linewidth=1, label='Imaginary Part of Shear Force' if element.id == elements[0].id else "")
            
            # Connect moment diagram back to the structure
            plt.plot([x[0], shearforce_x[0]], [z[0], shearforce_z[0]], 'b--', linewidth=1)
            plt.plot([x[-1], shearforce_x[-1]], [z[-1], shearforce_z[-1]], 'b--', linewidth=1)
            plt.plot([x[0], shearforce_x_imag[0]], [z[0], shearforce_z_imag[0]], 'g--', linewidth=1)
            plt.plot([x[-1], shearforce_z_imag[-1]], [z[-1], shearforce_z_imag[-1]], 'g--', linewidth=1)  
        # Add legend and show the plot
        plt.legend()
        plt.show()
 
    def Plotshearstresses(self, elements, stresses, scale=1e-9):
     
        plt.figure(figsize=(10, 8))
        plt.title('Shear Stress Plot')
        plt.axis('off')
    
        # Plot the original structure (nodes and elements)
        for node in [node for element in elements for node in element.nodes]:
            plt.scatter(node.x, node.z, color='red', marker='o', label='Node' if 'Node' not in plt.gca().get_legend_handles_labels()[1] else "")
            plt.text(node.x+0.02, node.z+0.02, f'{node.id}', fontsize=9, ha='right')
        for element in elements:
            x_values = [element.nodes[0].x, element.nodes[1].x]
            z_values = [element.nodes[0].z, element.nodes[1].z]
            plt.plot(x_values, z_values, 'k--', linewidth=2, label='Element' if 'Element' not in plt.gca().get_legend_handles_labels()[1] else "")
    
        ss_nodes_real = {}
        ss_nodes_imag = {}
    
        # Plot the displaced elements (real and imaginary parts)
        for element in elements:
            # Extract the original coordinates of the nodes
            x0 = element.nodes[0].x
            z0 = element.nodes[0].z
            x1 = element.nodes[1].x
            z1 = element.nodes[1].z
            
            # Get the displacement components (assuming u_elem has shape (6, num_points))
            sigma_elem = stresses[element.id]
            tau_real = np.real(sigma_elem[1, :])
            tau_imag = np.imag(sigma_elem[1, :])
            
            num_points = len(tau_real)
            # Create an array of points along the length of the element
            s = np.linspace(0, 1, num_points)
      
            # Calculate the original positions along the element's length
            x = x0 + (x1 - x0) * s
            z = z0 + (z1 - z0) * s
          # Compute perpendicular direction for offsetting the moment line
            dx = x1 - x0
            dz = z1 - z0
            length = np.sqrt(dx**2 + dz**2)
            perp_x = dz / length  # Unit normal in X
            perp_z = dx / length   # Unit normal in Z          
            
            shearstress_x = x - scale * tau_real * perp_x
            shearstress_z = z + scale * tau_real * perp_z
            
            shearstress_x_imag = x - scale * tau_imag * perp_x
            shearstress_z_imag = z + scale * tau_imag * perp_z
            
            # Plot the shear stress
            plt.plot(shearstress_x, shearstress_z, 'b-', linewidth=1, label='Real Part of Shear Stress' if element.id == elements[0].id else "")
            plt.plot(shearstress_x_imag, shearstress_z_imag, 'g--', linewidth=1, label='Imaginary Part of Shear Stress' if element.id == elements[0].id else "")
            
            # Connect moment diagram back to the structure
            plt.plot([x[0], shearstress_x[0]], [z[0], shearstress_z[0]], 'b--', linewidth=1)
            plt.plot([x[-1], shearstress_x[-1]], [z[-1], shearstress_z[-1]], 'b--', linewidth=1)
            plt.plot([x[0], shearstress_x_imag[0]], [z[0], shearstress_z_imag[0]], 'g--', linewidth=1)
            plt.plot([x[-1], shearstress_z_imag[-1]], [z[-1], shearstress_z_imag[-1]], 'g--', linewidth=1)  
        # Add legend and show the plot
        plt.legend()
        plt.show()

    def Plotaxialstresses(self, elements, stresses, scale=1e-9):
     
        plt.figure(figsize=(10, 8))
        plt.title('Axial Stress Plot')
        plt.axis('off')
    
        # Plot the original structure (nodes and elements)
        for node in [node for element in elements for node in element.nodes]:
            plt.scatter(node.x, node.z, color='red', marker='o', label='Node' if 'Node' not in plt.gca().get_legend_handles_labels()[1] else "")
            plt.text(node.x+0.02, node.z+0.02, f'{node.id}', fontsize=9, ha='right')
        for element in elements:
            x_values = [element.nodes[0].x, element.nodes[1].x]
            z_values = [element.nodes[0].z, element.nodes[1].z]
            plt.plot(x_values, z_values, 'k--', linewidth=2, label='Element' if 'Element' not in plt.gca().get_legend_handles_labels()[1] else "")
    
        as_nodes_real = {}
        as_nodes_imag = {}
    
        # Plot the displaced elements (real and imaginary parts)
        for element in elements:
            # Extract the original coordinates of the nodes
            x0 = element.nodes[0].x
            z0 = element.nodes[0].z
            x1 = element.nodes[1].x
            z1 = element.nodes[1].z
            
            # Get the displacement components (assuming u_elem has shape (6, num_points))
            sigma_elem = stresses[element.id]
            sigma_xx_real = np.real(sigma_elem[0, :])
            sigma_xx_imag = np.imag(sigma_elem[0, :])
            
            num_points = len(sigma_xx_real)
            # Create an array of points along the length of the element
            s = np.linspace(0, 1, num_points)
      
            # Calculate the original positions along the element's length
            x = x0 + (x1 - x0) * s
            z = z0 + (z1 - z0) * s
          # Compute perpendicular direction for offsetting the moment line
            dx = x1 - x0
            dz = z1 - z0
            length = np.sqrt(dx**2 + dz**2)
            perp_x = dz / length  # Unit normal in X
            perp_z = dx / length   # Unit normal in Z          
            
            axialstress_x = x + scale * sigma_xx_real * perp_x
            axialstress_z = z - scale * sigma_xx_real * perp_z
            
            axialstress_x_imag = x + scale * sigma_xx_imag * perp_x
            axialstress_z_imag = z + scale * sigma_xx_imag * perp_z
            
            # Plot the shear stress
            plt.plot(axialstress_x, axialstress_z, 'b-', linewidth=1, label='Real Part of Axial Stress' if element.id == elements[0].id else "")
            plt.plot(axialstress_x_imag, axialstress_z_imag, 'g--', linewidth=1, label='Imaginary Part of Axial Stress' if element.id == elements[0].id else "")
            
            # Connect moment diagram back to the structure
            plt.plot([x[0], axialstress_x[0]], [z[0], axialstress_z[0]], 'b--', linewidth=1)
            plt.plot([x[-1], axialstress_x[-1]], [z[-1], axialstress_z[-1]], 'b--', linewidth=1)
            plt.plot([x[0], axialstress_x_imag[0]], [z[0], axialstress_z_imag[0]], 'g--', linewidth=1)
            plt.plot([x[-1], axialstress_z_imag[-1]], [z[-1], axialstress_z_imag[-1]], 'g--', linewidth=1)  
        # Add legend and show the plot
        plt.legend()
        plt.show()

    def Plotbendingstresses(self, elements, stresses, scale=1e-9):
     
        plt.figure(figsize=(10, 8))
        plt.title('Bending Stress Plot')
        plt.xlabel('X Coordinate')
        plt.ylabel('Z Coordinate')
        plt.grid(True)
        plt.axis('equal')
    
        # Plot the original structure (nodes and elements)
        for node in [node for element in elements for node in element.nodes]:
            plt.scatter(node.x, node.z, color='red', marker='o', label='Node' if 'Node' not in plt.gca().get_legend_handles_labels()[1] else "")
            plt.text(node.x+0.02, node.z+0.02, f'{node.id}', fontsize=9, ha='right')
        for element in elements:
            x_values = [element.nodes[0].x, element.nodes[1].x]
            z_values = [element.nodes[0].z, element.nodes[1].z]
            plt.plot(x_values, z_values, 'k--', linewidth=2, label='Element' if 'Element' not in plt.gca().get_legend_handles_labels()[1] else "")
    
        as_nodes_real = {}
        as_nodes_imag = {}
    
        # Plot the displaced elements (real and imaginary parts)
        for element in elements:
            # Extract the original coordinates of the nodes
            x0 = element.nodes[0].x
            z0 = element.nodes[0].z
            x1 = element.nodes[1].x
            z1 = element.nodes[1].z
            
            # Get the displacement components (assuming u_elem has shape (6, num_points))
            sigma_elem = stresses[element.id]
            sigma_yy_real = np.real(sigma_elem[-1, :])
            sigma_yy_imag = np.imag(sigma_elem[-1, :])
            
            num_points = len(sigma_yy_real)
            # Create an array of points along the length of the element
            s = np.linspace(0, 1, num_points)
      
            # Calculate the original positions along the element's length
            x = x0 + (x1 - x0) * s
            z = z0 + (z1 - z0) * s
          # Compute perpendicular direction for offsetting the moment line
            dx = x1 - x0
            dz = z1 - z0
            length = np.sqrt(dx**2 + dz**2)
            perp_x = dz / length  # Unit normal in X
            perp_z = dx / length   # Unit normal in Z          
            
            bendingstress_x = x - scale * sigma_yy_real * perp_x
            bendingstress_z = z + scale * sigma_yy_real * perp_z
            
            bendingstress_x_imag = x - scale * sigma_yy_imag * perp_x
            bendingstress_z_imag = z + scale * sigma_yy_imag * perp_z
            
            # Plot the shear stress
            plt.plot(bendingstress_x, bendingstress_z, 'b-', linewidth=1, label='Real Part of Bending Stress' if element.id == elements[0].id else "")
            plt.plot(bendingstress_x_imag, bendingstress_z_imag, 'g--', linewidth=1, label='Imaginary Part of Bending Stress' if element.id == elements[0].id else "")
            
            # Connect moment diagram back to the structure
            plt.plot([x[0], bendingstress_x[0]], [z[0], bendingstress_z[0]], 'b--', linewidth=1)
            plt.plot([x[-1], bendingstress_x[-1]], [z[-1], bendingstress_z[-1]], 'b--', linewidth=1)
            plt.plot([x[0], bendingstress_x_imag[0]], [z[0], bendingstress_z_imag[0]], 'g--', linewidth=1)
            plt.plot([x[-1], bendingstress_z_imag[-1]], [z[-1], bendingstress_z_imag[-1]], 'g--', linewidth=1)  
        # Add legend and show the plot
        plt.legend()
        plt.show()

    def ShowStructure(self, title='Structure Configuration'):
        plt.title(title)
        plt.xlabel('X Coordinate')
        plt.ylabel('Z Coordinate')
        plt.legend()
        plt.grid(True)
        plt.axis('equal')
        plt.show()